# Copyright 2020 (c) Netguru S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from itertools import cycle
from typing import List, Dict

import matplotlib.pyplot as plt
import numpy as np

from opoca.evaluation.utils import calculate_roc_auc, calculate_prauc


class Plotter:
    """Class for plotting metrics for evaluating models."""
    def plot_roc_auc(self, y_true: np.ndarray, y_prob: np.ndarray, class_names: List[str] = None,
                     classes_to_plot: List[str] = None, show: bool = True, save_path: str = None):
        """
        Plots Receiver Operating Characteristic Area Under Curve.
        Parameters
        ----------
        y_true: np.ndarray
            Array of true labels (last shape of array should equal to number of classes).
        y_prob: np.ndarray
            Array of predicted probabilities (last shape of array should equal to number of classes).
        class_names: List of strings, optional
            Names of classes.
        classes_to_plot: List of strings, optional
            If provided, only the given classes get plotted.
            Otherwise all classes are plotted.
        show: bool, optional, default = True
            If True, the plot gets displayed.
        save_path: str, optional
            If provided, the plot gets saved in given path.
            The path's filename should end in desired save format e.g. 'bar_plot.png'.
        """
        class_names = class_names if class_names else [f'class_{i}' for i in range(y_prob.shape[-1])]
        fpr, tpr, roc_auc_dict = calculate_roc_auc(y_true, y_prob, class_names)
        self._plot_roc_auc(fpr, tpr, roc_auc_dict, class_names, classes_to_plot, show, save_path)

    def _plot_roc_auc(self, fpr: Dict[str, np.ndarray], tpr: Dict[str, np.ndarray], roc_auc_dict: Dict[str, np.ndarray],
                      class_names: List[str], classes_to_plot: List[str] = None, show: bool = True,
                      save_path: str = None):
        """
        Plots Receiver Operating Characteristic Area Under Curve.
        Parameters
        ----------
        fpr: dict
            Dictionary in which keys are names of classes and values are False Positives Ratio arrays.
        tpr: dict
            Dictionary in which keys are names of classes and values are True Positives Ratio arrays.
        roc_auc_dict: dict
            Dictionary in which keys are names of classes and values are calculated ROC area under curve.
        class_names: List of strings
            Names of classes.
        classes_to_plot: List of strings, optional
            If provided, only the given classes get plotted.
            Otherwise all classes are plotted.
        show: bool, optional, default = True
            If True, the plot gets displayed.
        save_path: str, optional
            If provided, the plot gets saved in given path.
            The path's filename should end in desired save format e.g. 'bar_plot.png'.
        """
        lw = 2
        colors = cycle(['turquoise', 'darkorange', 'cornflowerblue', 'teal', 'rosybrown',
                        'maroon', 'lawngreen', 'lightskyblue', 'palevioletred', 'indigo'])

        # Plot all ROC curves
        plt.figure(figsize=(24, 16))
        ax = plt.subplot(111)

        plt.plot(fpr["macro avg"], tpr["macro avg"],
                 label=f'macro-average ROC curve (area = {roc_auc_dict["macro avg"]:0.2f})',
                 color='navy', linestyle=':', linewidth=4)

        classes_to_plot = classes_to_plot if classes_to_plot else class_names

        for i, color in zip(classes_to_plot, colors):
            plt.plot(fpr[i], tpr[i], color=color, lw=lw,
                     label=f'ROC curve of class {i} (area = {roc_auc_dict[i]:0.2f})')

        plt.plot([0, 1], [0, 1], 'k--', lw=lw)
        plt.xlim([0.0, 1.0])
        plt.ylim([0.0, 1.05])
        plt.xlabel('False Positive Rate', {'fontsize': 15})
        plt.ylabel('True Positive Rate', {'fontsize': 15})
        plt.title('Receiver Operating Characteristic curve', {'fontsize': 20})
        plt.xticks(fontsize=12)
        plt.yticks(fontsize=12)
        box = ax.get_position()
        ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])
        ax.legend(loc='center left', bbox_to_anchor=(1, 0.5), prop=dict(size=14))
        if save_path:
            plt.savefig(save_path)
        if show:
            plt.show()

    def plot_prauc(self, y_true: np.ndarray, y_prob: np.ndarray, class_names: List[str] = None,
                   classes_to_plot: List[str] = None, show: bool = True, save_path: str = None):
        """
        Plots Precision-Recall Area Under Curve.
        Parameters
        ----------
        y_true: np.ndarray
            Array of true labels (last shape of array should equal to number of classes).
        y_prob: np.ndarray
            Array of predicted probabilities (last shape of array should equal to number of classes).
        class_names: List of strings, optional
            Names of classes.
        classes_to_plot: List of strings, optional
            If provided, only the given classes get plotted.
            Otherwise all classes are plotted.
        show: bool, optional, default = True
            If True, the plot gets displayed.
        save_path: str, optional
            If provided, the plot gets saved in given path.
            The path's filename should end in desired save format e.g. 'bar_plot.png'.
        """
        precision, recall, average_precision = calculate_prauc(y_true, y_prob)

        colors = cycle(['navy', 'turquoise', 'darkorange', 'cornflowerblue', 'teal', 'rosybrown',
                        'maroon', 'lawngreen', 'lightskyblue', 'palevioletred', 'indigo'])

        plt.figure(figsize=(24, 16))
        ax = plt.subplot(111)

        f_scores = np.linspace(0.2, 0.8, num=4)
        lines = []
        labels = []
        for f_score in f_scores:
            x = np.linspace(0.01, 1)
            y = f_score * x / (2 * x - f_score)
            l, = ax.plot(x[y >= 0], y[y >= 0], color='gray', alpha=0.2)
            ax.annotate(f'f1={f_score:0.1f}', xy=(0.9, y[45] + 0.02))

        lines.append(l)
        labels.append('iso-f1 curves')
        l, = ax.plot(recall["micro"], precision["micro"], color='gold', lw=2)
        lines.append(l)
        labels.append(f'Micro-average Precision-recall (area = {average_precision["micro"]:0.2f})')

        # get positions of classes_to_plot in class_names
        if classes_to_plot:
            if class_names:
                classes_to_plot_index = [i for i, name in enumerate(class_names) if name in classes_to_plot]
            else:
                classes_to_plot_index = [i for i in list(range(y_prob.shape[-1])) if i in classes_to_plot]
        else:
            classes_to_plot_index = list(range(y_prob.shape[-1]))

        for i, color in zip(classes_to_plot_index, colors):
            l, = ax.plot(recall[i], precision[i], color=color, lw=2)
            lines.append(l)
            if class_names:
                labels.append(
                    f'Precision-recall for class {class_names[i]} (area = {average_precision[i]:0.2f})')
            else:
                labels.append(f'Precision-recall for class {i} (area = {average_precision[i]:0.2f})')

        fig = plt.gcf()
        fig.subplots_adjust(bottom=0.25)
        plt.xlim([0.0, 1.0])
        plt.ylim([0.0, 1.05])
        plt.xlabel('Recall', {'fontsize': 15})
        plt.ylabel('Precision', {'fontsize': 15})
        plt.title('Precision-Recall curve', {'fontsize': 20})
        plt.xticks(fontsize=12)
        plt.yticks(fontsize=12)

        # Shrink current axis by 20%
        box = ax.get_position()
        ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])

        # Put a legend to the right of the current axis
        ax.legend(lines, labels, loc='center left', bbox_to_anchor=(1, 0.5), prop=dict(size=14))

        if save_path:
            plt.savefig(save_path)
        if show:
            plt.show()



# Copyright 2020 (c) Netguru S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import optuna
from horology import timed
from optuna import Trial
from optuna.trial import FixedTrial
from sklearn.datasets import load_breast_cancer
from sklearn.svm import SVC

from opoca.data.dataset import Dataset
from opoca.data.splitter import StratifiedSplitter
from opoca.hyperparam_search.objective import Objective
from opoca.hyperparam_search.pocstudy import POCStudy


class SimpleSVMObjective(Objective):
    """ Very simple objective that depends only on two parameters."""

    def create_model(self, trial: Trial):
        """
        In class that inherits from Objective, this method must be implemented
        and return a model. Parameters of the model should be obtained with the
        trial object.
        """
        C = trial.suggest_loguniform('C', 1e-6, 1e6)
        kernel = trial.suggest_categorical('kernel', ['linear', 'poly', 'rbf', 'sigmoid'])

        model = SVC(C=C, kernel=kernel)

        return model


class MultiLevelSVMObjective(Objective):
    """ More sophisticated example with non-constant dimension of search space"""

    def create_model(self, trial: Trial):
        """
        All optimized model parameters should be obtained form trial object
        using one of `suggest_*` methods. There is no requirement that always
        the same parameters are used. Here, depending on kernel, only
        applicable parameters are used. E.g. 'degree' make sense only for the
        polynomial kernel.
        """
        C = trial.suggest_loguniform('C', 1e-6, 1e6)
        dfs = trial.suggest_categorical('decision_function_shape', ['ovo', 'ovr'])
        kernel = trial.suggest_categorical('kernel', ['linear', 'poly', 'rbf', 'sigmoid'])
        max_iter = 1000

        if kernel == 'poly':
            degree = trial.suggest_int('degree', 2, 4)
            return SVC(C=C, kernel=kernel,
                       degree=degree,
                       decision_function_shape=dfs, max_iter=max_iter)

        elif kernel in ['rbf', 'sigmoid']:
            gamma = trial.suggest_categorical('gamma', ['scale', 'auto'])
            return SVC(C=C, kernel=kernel,
                       gamma=gamma,
                       decision_function_shape=dfs, max_iter=max_iter)

        # kernel == 'linear'
        return SVC(C=C, kernel=kernel,
                   decision_function_shape=dfs, max_iter=max_iter)


@timed
def example():
    """
    In this example we search for best hyperparameters for a Support Vector
    Machine Classifier on the breast cancer dataset.

    We split the data for training and testing (we do not need a separate
    validation set, because POCStudy internally does a cross validation).

    We run optimization for 60 seconds. You may see some ConvergenceWarning
    from SVC. Don't worry - those trials are usually pruned anyway. Number of
    explored points depends on how fast is your machine.

    We use the best params to train the model on the whole train set and score
    it on the test set.

    Finally we plot the optimization history. For first 10 trials, optuna will
    perform random search to warm up TPE, then it will focus on exploring most
    promising areas. Observe that after 15th iteration, multiple trials are
    pruned.
    """

    # Remove next 3 lines if you want to see some SVC warnings.
    import warnings
    from sklearn.exceptions import ConvergenceWarning
    warnings.filterwarnings(action='ignore', category=ConvergenceWarning)

    # Load and optionally split data.
    x, y = load_breast_cancer(return_X_y=True)
    dataset = Dataset(x, y, 'sklearn_breast_cancer')
    splitter = StratifiedSplitter(train_ratio=0.7, val_ratio=0.0, random_state=2020)
    split = splitter.split(dataset)

    # Instantiate objective with train data.
    # You can pass a list of scorers - the first one wll be used for
    # optimization and the rest just for monitoring
    objective = MultiLevelSVMObjective(split.train, scorer=['accuracy',
                                                            'f1',
                                                            'recall',
                                                            'precision'])

    # Create study and run optimization (internally it will use 5-fold CV)
    study = POCStudy(objective)
    study.optimize(timeout_seconds=60, num_trials=None)

    # Now you can use best trial params to create a model...
    best_params = study.best_trial.params
    print('Creating model with best params:', best_params)
    best_trial = FixedTrial(best_params)
    best_model = objective.create_model(best_trial)

    # ... which you can fit and score.
    best_model.fit(X=split.train.x, y=split.train.y)
    test_score = best_model.score(X=split.test.x, y=split.test.y)
    print(f'test_score: {test_score:.3g}')

    # Plot optimization history.
    print('Plot will hopefully open in a browser.')
    fig = optuna.visualization.plot_optimization_history(study)
    fig.show()
    print('Done.')


if __name__ == '__main__':
    example()



# Copyright 2020 (c) Netguru S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import List
from warnings import warn

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.preprocessing import StandardScaler


class BaseTransformer(BaseEstimator, TransformerMixin, metaclass=ABCMeta):
    """
    Base Feature Transformer class.

    Use this class when using scikit-learn trainers.
    It supports the built-in get_feature_names method of these trainers.

    IMPORTANT: When inheriting from this class remember to update self.feature_names with names of created features
               in the fit method of the inheriting Transformer.
    """
    features_names_: List[str]

    @abstractmethod
    def __init__(self):
        self.feature_names_ = None

    def fit(self, df: pd.DataFrame, *args) -> BaseTransformer:
        self.feature_names_ = df.columns
        return self

    def transform(self, x) -> pd.DataFrame:
        x = x[self.feature_names_]
        return x

    def get_feature_names(self) -> List[str]:
        return self.feature_names_


class ColumnsSelector(BaseTransformer):
    """
    Selects columns provided as constructor argument
    """

    def __init__(self, columns_names: List[str]):
        super().__init__()
        self.columns_names = columns_names

    def fit(self, x: pd.DataFrame, y: pd.DataFrame = None) -> ColumnsSelector:
        self.feature_names_ = self.columns_names
        return self


class ColumnDropper(BaseTransformer):
    """
    Drops columns provided as argument to constructor

    Parameters
        ----------
        columns_names
            List of columns names that shall be dropped
        inplace
            Determines whether dropping shall be done inplace
    """

    def __init__(self, columns_names: List[str], inplace: bool = False):
        super().__init__()
        self.inplace = inplace
        self.columns_to_drop_ = columns_names

    def fit(self, x: pd.DataFrame, y: pd.DataFrame = None) -> ColumnDropper:
        self.feature_names_ = list(set(x.columns) - set(self.columns_to_drop_))
        return self

    def transform(self, x: pd.DataFrame) -> pd.DataFrame:
        if self.inplace:
            x.drop(columns=self.columns_to_drop_, inplace=True, axis=1)
        else:
            x = x.drop(columns=self.columns_to_drop_, axis=1)
        return x


class NaNColumnDropper(ColumnDropper):
    """
    Drops columns that have more NaNs than threshold * number of elements in column.

    Parameters
    ----------
    threshold
        Determines what maximal fraction of NaNs we allow
    inplace
        Determines whether columns drop should be done inplace
    """

    def __init__(self, threshold: float = 0.3, inplace: bool = False):
        super().__init__([], inplace=inplace)
        self.threshold = threshold

    def fit(self, x: pd.DataFrame, y: pd.DataFrame = None) -> NaNColumnDropper:
        columns_nan_ratio = x.isnull().sum(axis=0) / len(x)
        self.columns_to_drop_ = x.columns[(columns_nan_ratio >= self.threshold)].to_list()
        self.features_names_ = list(set(x.columns) - set(self.columns_to_drop_))
        return self


class SingleValueColumnDropper(ColumnDropper):
    """
    Drops column that have only 1 value

    Parameters
    ----------
    inplace
        Determines whether columns drop should be done inplace
    """

    def __init__(self, inplace: bool = False):
        super().__init__([], inplace=inplace)

    def fit(self, x: pd.DataFrame, y: pd.DataFrame = None) -> SingleValueColumnDropper:
        for col in x.columns:
            if len(x[col].unique()) == 1:
                self.columns_to_drop_.append(col)
        self.feature_names_ = list(set(x.columns) - set(self.columns_to_drop_))
        return self


class ExtremelyDominantColumnDropper(ColumnDropper):
    """
    Drops columns that have value that extremely dominants other values within that column

    Parameters
    ----------
    threshold
        Determines maximal fraction of mode of given column
    inplace
        Determines whether columns drop should be done inplace
    """

    def __init__(self, threshold: float = 0.99, inplace: bool = False):

        super().__init__([], inplace=inplace)
        if threshold <= 0. or threshold >= 1.0:
            raise ValueError("threshold has to be within (0, 1)")
        self.threshold = threshold

    def fit(self, x: pd.DataFrame, y: pd.DataFrame = None) -> ExtremelyDominantColumnDropper:
        for c in x.columns:
            mode = x[c].mode()[0]
            ratio = x[c].value_counts()[mode] / len(x[c])
            if ratio > self.threshold:
                self.columns_to_drop_.append(c)
        self.feature_names_ = list(set(x.columns) - set(self.columns_to_drop_))
        return self


class NanReplacer(BaseTransformer):
    """
    Replaces predefined values with np.nan.
    By default NaN representations is `["nan", "None", "none", "NaN"]`
    """
    def __init__(self, to_replace: List[str] = None):
        super().__init__()
        if to_replace is None:
            self.to_replace = ["nan", "None", "none", "NaN", '']
        else:
            self.to_replace = to_replace

    def transform(self, x):
        return x.replace(to_replace=self.to_replace, value=np.nan)


class TypeSelector(BaseTransformer):
    """
    Selects columns which types belong to `types_list`

    Parameters
    ----------
    types_list
        List of types to be selected
    """

    def __init__(self, types_list: List[str]):
        super().__init__()
        self.types_list = types_list

    def fit(self, x: pd.DataFrame, y: pd.DataFrame = None) -> TypeSelector:
        self.feature_names_ = x.select_dtypes(include=self.types_list).columns
        return self


class NumericalFeatureSelector(TypeSelector):
    """
    Selects numerical features
    """
    NUMERICAL_TYPES = [
        "float",
        "float16",
        "float32",
        "float64",
        "int",
        "int8",
        "int16",
        "int32",
        "int64",
        "uint8",
        "uint16",
        "uint32",
        "uint64",
        "Int8",
        "Int16",
        "Int32",
        "Int64",
        "Uint8",
        "Uint16",
        "Uint32",
        "Uint64"
    ]

    def __init__(self):
        super().__init__(self.NUMERICAL_TYPES)


class IntegerFeatureSelector(TypeSelector):
    """
    Selects integer columns
    """
    INTEGER_TYPES = [
        "int",
        "int8",
        "int16",
        "int32",
        "int64",
        "uint8",
        "uint16",
        "uint32",
        "uint64"
    ]

    def __init__(self):
        super().__init__(self.INTEGER_TYPES)


class FloatFeaturesSelector(TypeSelector):
    """
    Selects float columns
    """
    FLOAT_TYPES = [
        "float",
        "float16",
        "float32",
        "float64"
    ]

    def __init__(self):
        super().__init__(self.FLOAT_TYPES)


class CategoricalFeatureSelector(TypeSelector):
    """
    Selects categorical columns
    """
    CATEGORY_TYPES = [
        "category",
        "object"
    ]

    def __init__(self):
        super().__init__(self.CATEGORY_TYPES)


class BooleanFeatureSelector(TypeSelector):
    """
    Selects boolean columns
    """

    def __init__(self):
        super().__init__(["bool"])


class TopKEncoder(BaseTransformer):
    """
    Encodes top k values in columns and rest as `OTHER_REPRESENTATION`. As a result there are k + 1 unique values in
    encoded columns.

    Important: works in place!!!

    Parameters
    ----------
    column_names
        List of columns to be encoded
    top_k
        Number of values to be left
    """
    OTHER_REPRESENTATION = "other"

    def __init__(self, column_names: List[str], top_k: int = 5):
        super().__init__()

        if top_k < 1:
            raise ValueError("top_k has to be greater than 0")

        self.column_names = column_names
        self.top_k = top_k
        self.values = {}
        self.cols_transformed = []

    def fit(self, x: pd.DataFrame, y: pd.DataFrame = None) -> TopKEncoder:
        if "float" in x[self.column_names].dtypes.unique():
            warn("Input data frame has floating columns. Those probably have only unique values and shouldn't "
                 "be encoded through this encoder")

        super(TopKEncoder, self).fit(x)

        for col in self.column_names:
            values = x[col].value_counts().index.tolist()
            if len(values) <= self.top_k:
                warn(f"top_k is greater than number of unique values in column {col}")
            if x[col].value_counts().values[self.top_k] == 1:
                warn("top k-th has only 1 element. Consider lower k")
            self.values[col] = values[:self.top_k]

        return self

    def transform(self, x: pd.DataFrame) -> pd.DataFrame:
        to_be_transformed = list(set(self.column_names) - set(self.cols_transformed))
        for col in to_be_transformed:
            other_indices = ~x[col].isin(self.values[col])
            nan_indices = x[col].isna()
            x.loc[other_indices, col] = self.OTHER_REPRESENTATION
            x.loc[nan_indices, col] = self.OTHER_REPRESENTATION
        return x


class UpperQuantileEncoder(BaseTransformer):
    """
    Encodes upper quantile values in column. This transformer computes cumulative sum of desceding sorted value counts
    for particular columns. Then, it divides them by total number of non empty values. Then it leaves (as they are )
    values that are lower than input quantile threshold and encodes rest to 'Other'.

    IMPORTANT: works in place !!!

    Example:
    >> df = pd.DataFrame({"col": [1, 1, 2, 1, 2, 0, 3, 7]})
    >> uqe = UpperQuantileEncoder(["col"], quantile_threshold=0.625)
    >> uqe.fit_transform(df)
         col
    0      1
    1      1
    2      2
    3      1
    4      2
    5  other
    6  other
    7  other

    Parameters
    ----------
    column_names
        Names of columns that shall be encoded
    quantile_threshold
        Determines which values to leave as they are (lower than quantile_threshold)
    """
    OTHER_REPRESENTATION = "other"

    def __init__(self, column_names: List[str], quantile_threshold: float = 0.5):
        super().__init__()
        if quantile_threshold <= 0 or quantile_threshold >= 1.0:
            raise ValueError("quantile_threshold has to be within (0, 1)")
        self.quantile_threshold = quantile_threshold
        self.column_names = column_names
        self.column_to_staying_values = {}

    def fit(self, x: pd.DataFrame, y: pd.DataFrame = None) -> UpperQuantileEncoder:
        if "float" in x[self.column_names].dtypes.unique():
            warn("Input data frame has floating columns. Those probably have only unique values and shouldn't "
                 "be encoded through this encoder")

        super(UpperQuantileEncoder, self).fit(x)

        for col in self.column_names:
            top_percentile_values = self.__get_top_percentile_values(x[col])
            if len(top_percentile_values) == 0:
                warn(f"quantile_threshold is too low. {col} won't be changed")
            else:
                self.column_to_staying_values[col] = top_percentile_values

        return self

    def transform(self, x: pd.DataFrame) -> pd.DataFrame:
        for col, staying_values in self.column_to_staying_values.items():
            other_indices = ~x[col].isin(staying_values)
            nan_indices = x[col].isna()
            x.loc[other_indices, col] = self.OTHER_REPRESENTATION
            x.loc[nan_indices, col] = self.OTHER_REPRESENTATION
        return x

    def __get_top_percentile_values(self, ps: pd.Series):
        total = ps.value_counts().sum()
        cumsum = ps.value_counts().sort_values(ascending=False).cumsum() / total
        return set(cumsum.loc[cumsum <= self.quantile_threshold].index)


class CurrencyTransformer(BaseTransformer):
    """
    Transforms columns with amount with respect to `CURRENCY_EXCHANGE_MULTIPLIER` dictionary.

    Parameters
    ----------
    amount_cols:
        List of amount columns to be transformed
    currency_col:
        Column that determines currency
    """
    CURRENCY_EXCHANGE_MULTIPLIER = {
        'PLN': 1,
        'EUR': 4,
        'USD': 4,
        'GBP': 6
    }

    def __init__(self, amount_cols: List[str], currency_col: str):
        super().__init__()
        self.currency_col = currency_col
        self.amount_cols = amount_cols

    def transform(self, x: pd.DataFrame) -> pd.DataFrame:
        multiplier = x[self.currency_col].map(self.CURRENCY_EXCHANGE_MULTIPLIER)
        for c in self.amount_cols:
            x[c] = x[c].multiply(multiplier, axis="index")
        return x


class LogTransformer(BaseTransformer):
    """
    Applies log to columns provided in constructor. It uses log(x+1) in order to make sure it does not explode for small
    values.

    Parameters
    ----------
    cols:
        list of columns that log transformation shall be applied to
    inplace:
        Determines whether log transformation is done inplace. If not, then separate columns are added for log
        transformation with suffixes `_log`
    """

    def __init__(self, cols: List = None, inplace: bool = True):
        super().__init__()
        self.in_place = inplace
        self.cols = cols

    def fit(self, x: pd.DataFrame, y: pd.DataFrame = None) -> LogTransformer:
        if self.cols is None:
            self.cols = x.columns
        return self

    def transform(self, x: pd.DataFrame) -> pd.DataFrame:
        if self.in_place:
            cols = self.cols
        else:
            cols = [f'{col}_log' for col in self.cols]
        x[cols] = np.log1p(x[self.cols])
        return x


class Scaler(BaseTransformer):
    """
    Wraps around `StandardScaler` so it provides `get_feature_names`, needed in complex pipelines

    Parameters
    ----------
    copy:
        If False, try to avoid a copy and do inplace scaling instead.
        This is not guaranteed to always work inplace; e.g. if the data is
        not a NumPy array or scipy.sparse CSR matrix, a copy may still be
        returned.

    with_mean:
        If True, center the data before scaling.
        This does not work (and will raise an exception) when attempted on
        sparse matrices, because centering them entails building a dense
        matrix which in common use cases is likely to be too large to fit in
        memory.

    with_std:
        If True, scale the data to unit variance (or equivalently,
        unit standard deviation).
    """

    def __init__(self, copy: bool = True, with_mean: bool = True, with_std: bool = True):
        super().__init__()
        self.scaler = StandardScaler(copy=copy, with_mean=with_mean, with_std=with_std)

    def fit(self, x: pd.DataFrame, y: pd.DataFrame = None) -> Scaler:
        self.scaler.fit(x)
        self.feature_names_ = x.columns.tolist()
        return self

    def transform(self, x: pd.DataFrame) -> np.ndarray:
        return self.scaler.transform(x)


class ScaleByMean(BaseTransformer):
    """
    Scales feature `numerical_feature` by mean according to grouping by categorical or boolean feature `group_feature`
    """

    def __init__(self, numerical_feature: str, group_feature: str):
        """

        Parameters
        ----------
        numerical_feature
            Numerical feature name that transformation is applied to
        group_feature
            Categorical or boolean feature name that is used for aggregation
        """

        super().__init__()
        self.numerical_feature = numerical_feature
        self.group_feature = group_feature
        self.name = f'{self.numerical_feature}_{self.group_feature}_scale_by_mean'

    def fit(self, x: pd.DataFrame, y: pd.DataFrame = None) -> ScaleByMean:
        self.feature_names_ = x.columns.tolist()
        self.feature_names_.append(self.name)
        return self

    def transform(self, x: pd.DataFrame) -> pd.DataFrame:
        x[self.name] = x.groupby(self.group_feature)[self.numerical_feature].apply(lambda x: x / (np.mean(x) + 1.e-6))
        return x



# Copyright 2020 (c) Netguru S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
from functools import lru_cache
from typing import Union, List, Tuple, Optional

import pandas as pd
import quilt3
from dynaconf import settings
from quilt3 import list_package_versions

from opoca.data.dataset import Dataset
from opoca.data.split import Split, SPLIT_FLAG_NAME

# Following default setting may be overridden by exporting env variables with prefix DYNACONF_, e.g.
# $ export DYNACONF_QUILT_DATA_DIR=".data/quilt"
# or by using env file (see docs of dynaconf)
QUILT_DATA_DIR = settings.get('QUILT_DATA_DIR', default='~/.quilt_data/')
S3_BUCKET = settings.get('S3_BUCKET')
QUILT_URL = settings.get('QUILT_URL')

SPLIT_TABLE_SUFFIX = '__split_table.csv'


def set_up_quilt() -> str:
    """ Set up default values in the quilt config

    Assures that QUILT_DATA_DIR folder exists and set it in quilt3 config as
    the default install location.

    Returns
    -------
    default install location: str
        Local path where packages will be downloaded
    """
    quilt3.config(QUILT_URL)
    quilt3.config(default_remote_registry=S3_BUCKET)
    default_install_location = os.path.expanduser(QUILT_DATA_DIR)
    os.makedirs(default_install_location, exist_ok=True)
    quilt3.config(default_install_location='file://' + default_install_location)

    return default_install_location


def pd_read_any(filename: str, **kwargs) -> pd.DataFrame:
    """ Wrapper around pandas read_*

    Tries to infer proper read method by file extension.

    Parameters
    ----------
    filename: str
        Path to a file with tabular data that should be read
    **kwargs:
        kwargs that are further passed to the pandas read method

    Returns
    -------
    data: pd.DataFrame
        loaded tabular data

    Raises
    ------
    ValueError: if file extensions in unknown.
    """
    data_format = filename.split('.')[-1].lower().strip()

    if data_format == 'csv':
        return pd.read_csv(filename, **kwargs)

    if data_format in ['feather', 'ftr', ]:
        return pd.read_feather(filename, **kwargs)

    if data_format == 'parquet':
        return pd.read_parquet(filename, **kwargs)

    if data_format in ['pickle', 'pkl']:
        return pd.read_pickle(filename, **kwargs)

    raise ValueError(f'Unknown data format: {data_format} of file {filename}')


def pd_save_any(df: Union[pd.Series, pd.DataFrame], filename: str, **kwargs):
    """ Wrapper around pandas to_*

    Tries to infer proper save method by file extension.

    Parameters
    ----------
    df: pd.DataFrame
        Data to be saved
    filename: str
        Path to a file where data will be saved
    **kwargs:
        kwargs that are further passed to the pandas to_ method if it accepts any

    Raises
    ------
    ValueError: if file extensions in unknown.
    """
    data_format = filename.split('.')[-1].lower().strip()

    if data_format == 'csv':
        df.to_csv(filename, **kwargs)

    elif data_format in ['feather', 'ftr', ]:
        df.to_feather(filename)

    elif data_format == 'parquet':
        df.to_parquet(filename, **kwargs)

    elif data_format in ['pickle', 'pkl']:
        df.to_pickle(filename, **kwargs)

    else:
        raise ValueError(f'Unknown data format: {data_format} of file {filename}')


class DataHandler:
    """ Base class for data loaders that are specific for each dataset

    Parameters
    ----------
    dataset_name: str
        in quilt3 refereed to as `namespace`
    data_form: str
        in quilt3 referred to as `packagename`, it can take values like:
        'zipped', 'raw', 'preprocessed' etc, depending what forms of data are
        handled in _load_* methods of subclasses.
    """

    def __init__(self, dataset_name: str, data_form: str):
        self.package_name = f'{dataset_name}/{data_form}'
        self.package: Optional[quilt3.Package] = None
        self.default_data_file = None
        self.target_column_names = None
        self.install_location = set_up_quilt()

    def __repr__(self):
        return f'{self.__class__.__name__} for {self.package_name} (local data path: {self.local_data_dir})'

    @property
    def local_package_root(self):
        return os.path.join(self.install_location, self.package_name)

    @property
    def local_data_dir(self):
        return os.path.join(self.local_package_root, 'data')

    def list(self) -> List[Tuple[str, str]]:
        """ List package revisions

        Returns
        -------
        revisions: List[Tuple[str, str]]
            A list of two-element tuples (timestamp, hash).
            The timestamp is in a form of an epoch.
            The last entry instead of a timestamp has a word 'latest' and the
            same hash as the last but one entry.
        """
        return list(list_package_versions(self.package_name, registry=S3_BUCKET))

    @lru_cache(maxsize=16)
    def load(self, filename: Optional[str] = None, return_as: str = 'pandas',
             num_samples: Optional[int] = None, **kwargs) \
            -> Union[pd.Series, pd.DataFrame, Dataset, Split]:
        """ Load data set as a pandas DataFrame

        It assures that data is downloaded and loaded only once.

        Parameters
        ----------
        filename: str, optional
            If None, default_data_file from package metadata is used. If
            given, it tries to read this file from the package. Useful when
            there are additional files with data.
        return_as: str, optional
            If 'pandas', returns loaded data as a single pandas DateFrame or
            Series.
            If 'dataset', uses target_column_names from package metadata, to
            split for X and y, and packs it in a Dataset object.
            If 'split', loads split table from a dedicated file and returns as
            a Split object.
        num_samples: int or None, optional
            In None, all samples are returned (default).
            If an int, randomly chosen subset of data will be returned. Useful
            for big data sets or for unit-testing.
        **kwargs:
            kwargs that are further passed to the pandas read method.

        Returns
        -------
        df: pd.Series, pd.DataFrame, Dataset or Split
            loaded data

        Notes
        -----
        `lru_cache` decorator caches results in case of multiple calls.

        Examples
        --------
        >>> loader = DataHandler('playground', 'normalized')
        >>> dataset = loader.load(return_as='dataset')
        ...
        >>> dataset.x.describe()
        ...
        """
        return_as = return_as.lower()
        self._install()
        local_path = self._local_path_from_filename(filename)
        data = pd_read_any(local_path, **kwargs)

        if num_samples is not None:
            data = data.sample(n=num_samples)

        if return_as == 'pandas':
            return data

        y = pd.concat([data.pop(c) for c in self.target_column_names], 1)
        dataset = Dataset(x=data, y=y,
                          name=self.package_name,
                          meta_data=self.package.meta)

        if return_as == 'dataset':
            return dataset

        if return_as == 'split':
            split_table = self._load_split_table(local_path)
            return Split.from_split_table(dataset=dataset, split_table=split_table)

        raise ValueError(f'Unrecognized `return_as` value: {return_as}.')

    def save(self, df: pd.DataFrame, filename: Optional[str] = None,
             split_table: Optional[pd.Series] = None, **kwargs) -> str:
        """ Save data set as a pandas DataFrame

        It infers a proper format from the file name extension.

        Parameters
        ----------
        df: pd.DataFrame
            Data to be saved.
        filename: str, optional
            If None, default_data_file from package metadata is used and file
            is overwritten. If given, a new file is created.
            When creating a new package, it must be not None, because no
            metadata is present.
        split_table: pd.Series, optional
            Series with flags indicating to which set each row belongs.
            If not None, a split table will be saved in the same location under
            filename constructed by appending `SPLIT_TABLE_SUFFIX` to the main
            data file. The split table is always saved as csv.
        **kwargs:
            kwargs that are further passed to the pandas to_ method.

        Returns
        -------
        local_path: str
            Path to the file where data was written.
        """
        self._install()
        local_path = self._local_path_from_filename(filename)
        pd_save_any(df, local_path, **kwargs)

        if split_table is not None:
            split_local_path = local_path + SPLIT_TABLE_SUFFIX
            pd_save_any(split_table, split_local_path, index=True, header=False)

        return local_path

    def init(self):
        """ Init a new data package

        Should be used only when creating a new dataset or data form.

        Raises
        ------
        OSError
            when package with this names already exists.

        """
        os.makedirs(self.local_data_dir, exist_ok=True)
        self.package = quilt3.Package()
        self.package.set_dir('/', self.local_package_root)

    def push(self, message='', new_meta: Optional[dict] = None):
        """ Commit and push the current content of local_package_root to the
        remote quilt server

        Parameters
        ----------
        message: str, optional
            A commit message visible in the quilt server.
        new_meta: dictionary, optional
            A dictionary with new metadata for the package.
            If None, existing metadata is used (if any).
            When pushing a new package, it is good to set at least
            `default_data_file` and `target_column_names`.
        """
        if new_meta is not None:
            meta = new_meta
        elif self.package.meta is not None:
            meta = self.package.meta.copy()
        else:
            meta = {}

        self.package.set_dir('/', self.local_package_root)
        # setting a dir removes all metadata, so we have to restore it here:
        self.package.set_meta(meta)
        self.package.push(self.package_name, message=message)
        self.package = None

    def _install(self):
        """ Download latest version of a dataset if needed

        This method is called by multiple methods. It uses quilt3 `install`
        method to download the latest version of the dataset. It avoids
        downloading data if it is already present in the default location.
        It also extracts some metadata from the package.

        """
        if self.package is not None:
            return

        quilt3.Package.install(self.package_name)
        # Following is a workaround for lack of `return` in the install
        # method in quilt 3.1.8
        self.package = quilt3.Package.browse(self.package_name)
        self.default_data_file = self.package.meta.get('default_data_file',
                                                       None)
        self.target_column_names = self.package.meta.get('target_column_names',
                                                         [])

    def _local_path_from_filename(self, filename: str):
        """ Finds an absolute local path for a given data file

        Parameters
        ----------
        filename: str or None
            If None, filename is read from package metadata.
            Keep in mind that filename may not contain '/' - nesting is not
            supported.

        Raises
        ------
        ValueError
            when filename is None and default_data_file is not specified for
            the package.
        """
        if filename is None:
            filename = self.default_data_file
            if filename is None:
                raise ValueError(f'No default_data_file for package '
                                 f'{self.package_name}. You must specify '
                                 f'filename explicitly.')

        return os.path.join(self.local_data_dir, filename)

    def _load_split_table(self, local_path: str) -> pd.Series:
        """ Load split table from a file

        Parameters
        ----------
        local_path: str
            Path to a base file with data for which a split table should be
            loaded.

        Returns
        -------
        split_table: pd.Series
            Series with int values indicating to which dataset a given row
            belongs. See `SplitFlag` enum for details.
        """
        local_path = local_path + SPLIT_TABLE_SUFFIX
        try:
            split_table = self.load(local_path, index_col=0, header=None,
                                    squeeze=True)
            return split_table.rename(SPLIT_FLAG_NAME)
        except FileNotFoundError:
            raise FileNotFoundError(f'Cannot find a split table file. It should'
                                    f' be named as the base file with data with'
                                    f' a suffix: {local_path}.')



# Copyright 2020 (c) Netguru S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Scoring machine learning models
"""
from collections import defaultdict
from typing import List

import numpy as np
import pandas as pd
from imblearn.pipeline import Pipeline
from sklearn.metrics import accuracy_score, classification_report

from opoca.data.dataset import Dataset
from opoca.evaluation.plotter import Plotter
from opoca.evaluation.utils import convert_list_probas_to_array, calculate_roc_auc, check_task


class Scorer:
    """Class for scoring models and plotting metrics."""
    def __init__(self, metrics: List[str] = None, threshold: float = 0.5, report: bool = False):
        """
        Parameters
        ----------
        metrics: List of strings, optional
            List of metrics' names as strings. If None, all supported metrics get calculated.
            Supported metrics: ['precision', 'recall', 'f1', 'accuracy', 'auc']
        threshold: float, optional
            Threshold for rounding prediction in case of binary classification
        report: bool, optional
            Determines whether to print classification report
        """
        self.report = report
        self.threshold = threshold
        self.metrics = ['precision', 'recall', 'f1', 'accuracy', 'auc'] if not metrics else metrics
        self.n_classes = None
        self.task = None
        self.fpr = None
        self.tpr = None
        self.roc_auc_dict = None
        self.class_names = None
        self.scores_dict = defaultdict(dict)

    def score(self, pipeline: Pipeline, dataset: Dataset, class_names: List[str] = None):
        """
        Computes scores for metrics provided in Scorer constructor. If y_true is multi class then scorer does
        macro average mode for precision/recall/f1. If y_true is multilabel the scores performs macro averaging.

        Parameters
        ----------
        pipeline: Pipeline
            Complete pipeline including features pipeline and classifier.
        dataset: Dataset
            Dataset containing x and y pd.DataFrames
            For Multiclass the shape of y_true should be 1-D (NOT one-hot encoded).
            For Multilabel the shape should be n-dimensional (where n is number of classes).
        class_names: List of strings, optional
            If given, the scores for separate classes will be displayed with appropriate names.

        Returns
        -------
        metrics: Dict
            Dictionary with metrics' names as keys and scores as values
        """

        x, y_true = dataset.x, dataset.y.to_numpy()

        if len(y_true.shape) == 1:
            y_true = y_true.reshape(-1, 1)

        # run inference for probabilities
        probabilities = pipeline.predict_proba(x)

        # check if the output of inference is a list (sklearn models often output probabilities in this form
        # in case of multilabel task). If yes, convert to a single array.
        if isinstance(probabilities, list):
            probabilities = convert_list_probas_to_array(probabilities)

        # check task type based on the true and predicted arrays.
        self.task = check_task(y_true, probabilities)

        # turn probabilities into predictions with chosen threshold
        if self.task in ['binary', 'multiclass']:
            predictions = np.argmax(probabilities, axis=-1)
        else:
            predictions = np.where(probabilities >= self.threshold, 1, 0)

        # assign number of classes based on given array
        self.n_classes = probabilities.shape[-1]

        # assign names of classes
        self.class_names = class_names if class_names else [f'class_{i}' for i in range(self.n_classes)]

        # check if any of ['precision', 'recall', 'f1', 'accuracy'] are in the metrics.
        # if yes generate classification report - it calculates all of these metrics.
        # it does not calculate accuracy for multilabel problem so additional check is done in such case.
        if [metric for metric in ['precision', 'recall', 'f1', 'accuracy'] if metric in self.metrics]:
            self.scores_dict.update(classification_report(y_true, predictions,
                                                          target_names=self.class_names, output_dict=True))
            if self.task == 'multilabel':
                self.scores_dict['accuracy'] = accuracy_score(y_true, predictions)

        if 'auc' in self.metrics:
            self.fpr, self.tpr, self.roc_auc_dict = calculate_roc_auc(y_true, probabilities,
                                                                      self.class_names, self.task)
            for key, value in self.roc_auc_dict.items():
                self.scores_dict[key]['auc'] = value

        if self.report:
            print(pd.DataFrame(self.scores_dict).transpose())

        return self._get_metrics()

    def _get_metrics(self, avg_type: str = 'macro avg') -> dict:
        """
        Convenience method extracting proper metrics from the score_dict
        and outputting them in the right format.

        Parameters
        ----------
        avg_type: str, default = 'macro avg'
            Type of average to use as the output value for chosen metrics.
            Possible averages are: ['macro avg', 'micro avg', 'weighted avg']

        Returns
        -------
        out: dict
            Dictionary, in which keys are names of metrics, and values are calculated metrics scores.
        """
        scores = dict()
        for metric in self.metrics:
            if metric == 'accuracy':
                scores.update({'accuracy': self.scores_dict['accuracy']})
            elif metric == 'f1':
                scores.update({'f1': self.scores_dict[avg_type]['f1-score']})
            else:
                scores.update({metric: self.scores_dict[avg_type][metric]})
        return scores

    def plot_roc_auc(self, classes_to_plot: List[str] = None, show: bool = True):
        """
        Plots Receiver Operating Characteristic Area Under Curve.
        The plot is partially based on previous calculations performed in the `score` method.

        IMPORTANT: Please run the `score` method with 'auc' metric inside of `metrics` argument before
        using this method.
        Parameters
        ----------
        classes_to_plot: List of strings, optional
            If provided, only the given classes get plotted.
            Otherwise all classes are plotted.
        show: bool, optional, default = True
            If True, the plot gets displayed.
        """
        if not self.fpr:
            raise RuntimeError("plot_roc_auc method called before running the `score` method with 'auc' in metrics.")
        plotter = Plotter()
        plotter._plot_roc_auc(self.fpr, self.tpr, self.roc_auc_dict, self.class_names, classes_to_plot, show=show)



# Copyright 2020 (c) Netguru S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from abc import abstractmethod, ABC
from math import sqrt
from typing import Optional
from warnings import warn

import numpy as np
import pandas as pd
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import train_test_split, StratifiedKFold

from opoca.data.adversarial_validation import create_adversarial_validation_data
from opoca.data.dataset import Dataset
from opoca.data.split import Split


class Splitter(ABC):
    @abstractmethod
    def split(self, data: Dataset) -> Split:
        pass


class RandomSplitter(Splitter):
    """
    Splits data randomly (uniformly) into train, val and test datasets. If `shuffle` kwarg is set to True,
    then it shuffles data before splitting.
    """

    def __init__(self, train_ratio: float = 0.6, val_ratio: float = 0.2,
                 shuffle: bool = True, random_state: Optional[int] = None):
        self.shuffle = shuffle
        self.train_ratio = train_ratio
        self.val_ratio = val_ratio
        self.test_ratio = 1. - train_ratio - val_ratio
        self.random_state = random_state

        assert self.train_ratio > 0
        assert self.val_ratio >= 0
        assert self.test_ratio > 0

    def split(self, data: Dataset) -> Split:
        x_train, x_test_val, y_train, y_test_val = self._split(data.x, data.y, self.train_ratio)
        train_dataset = Dataset(x=x_train, y=y_train, name=f"{data.name}/train")

        if self.val_ratio > 0:
            test_size = self.test_ratio / (self.val_ratio + self.test_ratio)
            x_test, x_val, y_test, y_val = self._split(x_test_val, y_test_val, test_size)
            val_dataset = Dataset(x=x_val, y=y_val, name=f"{data.name}/val")
            test_dataset = Dataset(x=x_test, y=y_test, name=f"{data.name}/test")
        else:  # self.val_ratio == 0
            val_dataset = None
            test_dataset = Dataset(x=x_test_val, y=y_test_val, name=f"{data.name}/test")

        return Split(train=train_dataset, val=val_dataset, test=test_dataset)

    def _split(self, x: pd.DataFrame, y: pd.DataFrame, train_size: float):
        return train_test_split(x, y, train_size=train_size, stratify=None,
                                shuffle=self.shuffle, random_state=self.random_state)


class StratifiedSplitter(RandomSplitter):
    """
    Performs stratified (with the same distribution of target variable) split into train, val and test datasets.
    Shuffles the data before performing the split.
    """

    def __init__(self, train_ratio: float = 0.6, val_ratio: float = 0.2, random_state: Optional[int] = None):
        super().__init__(train_ratio, val_ratio, shuffle=True, random_state=random_state)

    def _split(self, x: pd.DataFrame, y: pd.DataFrame, train_size: float):
        return train_test_split(x, y, train_size=train_size, stratify=y,
                                shuffle=self.shuffle, random_state=self.random_state)


class TimeSplitter(Splitter, ABC):
    """ Abstract base class for time splitters.

    Implements `split` method as a template method which does a few checks
    necessary to assure that the time split is reliable. In subclasses, one
    should implement `_split` method.
    """

    def split(self, dataset: Dataset) -> Split:
        """ Split dataset for three subsets: for training, validation and
        testing.

        Parameters
        ----------
        dataset: Dataset
            A time series Dataset. Its index must be a DatetimeIndex.

        Returns
        -------
        split: Split
            A split with strict time separation: the latest observation from the
            train set is guaranteed to be earlier than earliest observation
            from the validation set.
            Analogically for validation/test sets.
        """
        assert dataset.x.index.is_all_dates and dataset.y.index.is_all_dates, \
            'DataFrames index must have DatetimeIndex format.'

        x = dataset.x.sort_index()
        y = dataset.y.sort_index()

        assert x.index.equals(y.index), 'x and y must have the same index.'

        x_train, y_train, x_val, y_val, x_test, y_test = self._split(x, y)

        train_dataset = Dataset(x=x_train, y=y_train, name=f"{dataset.name}/train")
        val_dataset = Dataset(x=x_val, y=y_val, name=f"{dataset.name}/val")
        test_dataset = Dataset(x=x_test, y=y_test, name=f"{dataset.name}/test")

        return Split(train=train_dataset, val=val_dataset, test=test_dataset)

    @abstractmethod
    def _split(self, x, y) -> tuple:
        pass


class ProportionalTimeSplitter(TimeSplitter):
    """ Time split based on proportions of dataset

    Parameters
    ----------
    train_ratio: float, optional
        fraction of examples to be put in the training set
    val_ratio: float, optional
        fraction of examples to be put in the validation set

    Fraction of examples in the test set is calculated from the above ones.

    See Also
    --------
    `TimeSplitter` class docs.
    """

    def __init__(self, train_ratio: float = 0.6, val_ratio: float = 0.2):
        self.train_ratio = train_ratio
        self.val_ratio = val_ratio
        self.test_ratio = 1. - train_ratio - val_ratio

        assert self.train_ratio > 0
        assert self.val_ratio > 0
        assert self.test_ratio > 0

    def _split(self, x, y):
        train_val_break_ind = int(self.train_ratio * len(y.index))
        val_test_break_ind = int((self.train_ratio + self.val_ratio) * len(y.index))

        x_train, y_train = x.iloc[:train_val_break_ind], y.iloc[:train_val_break_ind]
        x_val, y_val = x.iloc[train_val_break_ind:val_test_break_ind], y.iloc[train_val_break_ind:val_test_break_ind]
        x_test, y_test = x.iloc[val_test_break_ind:], y.iloc[val_test_break_ind:]

        return x_train, y_train, x_val, y_val, x_test, y_test


class CutOffDateSplitter(TimeSplitter):
    """ Time split based on explicit cut-off dates

    Parameters
    ----------
    train_cutoff_date: str
        when training ends and validation begins
    val_cutoff_date: str
        when validation ends and testing begins

    `train_cutoff_date` must be earlier than `val_cutoff_date`.

    See Also
    --------
    `TimeSplitter` class docs.
    """

    def __init__(self, train_cutoff_date: str, val_cutoff_date: str):
        self.train_cutoff_date = pd.Timestamp(train_cutoff_date)
        self.val_cutoff_date = pd.Timestamp(val_cutoff_date)

        assert self.train_cutoff_date < self.val_cutoff_date, \
            'Train cut-off date must be earlier than val cut-off'

    def _split(self, x, y):
        x_train = x.loc[:self.train_cutoff_date]
        y_train = y.loc[:self.train_cutoff_date]
        x_val = x.loc[self.train_cutoff_date:self.val_cutoff_date]
        y_val = y.loc[self.train_cutoff_date:self.val_cutoff_date]
        x_test = x.loc[self.val_cutoff_date:]
        y_test = y.loc[self.val_cutoff_date:]

        # If the cutoff date is exactly equal to a timestamp in the index,
        # we may end up having one row in two sets,
        # because loc indexing is inclusive on both sides.
        last_train = x_train.tail(1).index
        first_val = x_val.head(1).index
        if last_train == first_val:
            x_train = x_train.iloc[:-1]
            y_train = y_train.iloc[:-1]

        last_val = x_val.tail(1).index
        first_test = x_test.head(1).index
        if last_val == first_test:
            x_val = x_val.iloc[:-1]
            y_val = y_val.iloc[:-1]

        return x_train, y_train, x_val, y_val, x_test, y_test


class AdversarialSplitter(Splitter):
    """ Split the train set into train and validation sets in such a way that
    the validation set is as much similar to the test set as possible.

    Useful when the test set is from different distribution then the train
    set. Use the `AdversarialValidator` to check if this is the case.

    Consider two scenarios:
    1. You have a train and test sets with labels. Then you simply want to get
    a subset of the train set to validate your model frequently on data similar
    to the test set, because you do not want to overfit on the test set and use
    it only to the final evaluation.
    2. Your test set has no labels. Then you have to chose a subset of the train
    set somehow to evaluate the model. `AdversarialSplitter` picks those samples
    that look like from the test set.

    Both scenarios are supported.

    Parameters
    ----------
    train: pd.DataFrame
        Test dataset is used for adversarial split fitting and is put in the
        final Split, divided into `split.train` and `split.val`.
    test: pd.DataFrame
        Test dataset is used for adversarial split fitting and is put in the
        final Split as the `split.test`. May not contain the columns which are
        defined in `target_columns`.
    target_columns: list of strings
        Names of columns that should become targets in the final split.
    val_ratio: float or None, optional
        Defines fraction of the train set that should be put in the validation
        set. If None, number of samples in the validation set will be same as
        in the test set, but no more than 50% of the train set.
    num_cv_splits: int, optional
        number of cv splits

    Remarks
    _______
    In cross validation procedure, multiple classifiers are created. They are
    then used to predict probabilities which can be interpreted how a given
    sample is likely to come from the test set. But each classifier may have
    different calibration, so comparing probabilities between elements from
    different folds may not be correct. Read more about calibration here:
    https://scikit-learn.org/stable/modules/calibration.html
    """

    def __init__(self, train: pd.DataFrame, test: pd.DataFrame,
                 target_columns: list,
                 val_ratio: Optional[float] = None, num_cv_splits=5):

        self.train = train.copy()
        self.test = test.copy()
        self.target_columns = target_columns

        self.x, self.y = create_adversarial_validation_data(train, test)

        if val_ratio is None:
            self.num_val = int(min(len(test), 0.5 * len(train)))
        else:
            self.num_val = int(val_ratio * len(train))

        self.num_cv_splits = num_cv_splits

    def split(self, data: Dataset) -> Split:
        """ Perform the adversarial split

        Parameters
        ----------
        data: Dataset
            Used only to get the name of the dataset. The data inside is not
            used, because the necessary data is provided in the constructor.
            Needed to follow the `Splitter` protocol.
            In order to avoid the warning, use it with a dummy Dataset like
            this:
            >>> split = adversarial_splitter.split(Dataset(None, None, 'pima'))

        Returns
        -------
        split: Split
            Split object where `train` is divided into `split.train` and
            `split.val`, and `test` is `split.test`.

        """
        if data.x is not None:
            warn('Actual data in `data` is not used. Only data.name is read.')

        cross_validation = StratifiedKFold(n_splits=self.num_cv_splits, shuffle=True)

        # this highly confusing heuristics limits number of trees in random
        # forest for small sets
        num_estimators = int(max(3, min(100, 0.5 * sqrt(len(self.y)))))

        testlike_probas = - np.ones_like(self.y, dtype='float')

        for cv_train_index, cv_val_index in cross_validation.split(self.x, self.y):
            x_train, x_val = self.x.iloc[cv_train_index], self.x.iloc[cv_val_index]
            y_train, _ = self.y.iloc[cv_train_index], self.y.iloc[cv_val_index]

            classifier = RandomForestClassifier(num_estimators, n_jobs=-1)
            classifier.fit(x_train, y_train)

            y_pred = classifier.predict_proba(x_val)[:, 1]

            testlike_probas[cv_val_index] = y_pred

        assert np.all(testlike_probas >= 0), 'Some samples were not classified.'

        y_with_predictions = self.y.copy().to_frame()
        y_with_predictions['testlike'] = testlike_probas

        y_with_predictions_from_train_only: pd.DataFrame = y_with_predictions.loc['train']
        # This removes the train/test level from MultiIndex, so further we deal
        # with a simple index
        assert np.all(y_with_predictions_from_train_only['is_test'] == 0), \
            'Mismatch in train/test datasets.'

        y_most_testlikes = y_with_predictions_from_train_only.sort_values('testlike')
        y_most_testlikes = y_most_testlikes.tail(self.num_val)
        # We don't really care about `y_most_testlikes` values anymore, we use
        # only its index to chose samples from primary train dataset.

        new_val = self.train[self.train.index.isin(y_most_testlikes.index)]
        new_train = self.train[~self.train.index.isin(y_most_testlikes.index)]

        train_dataset = Dataset.from_target_columns(new_train, self.target_columns,
                                                    name=data.name + '/train')
        val_dataset = Dataset.from_target_columns(new_val, self.target_columns,
                                                  name=data.name + '/val')
        test_dataset = Dataset.from_target_columns(self.test, self.target_columns,
                                                   name=data.name + '/test', allow_empty_y=True)

        return Split(train_dataset, val_dataset, test_dataset)



# Copyright 2020 (c) Netguru S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, List
from warnings import warn

import pandas as pd


@dataclass
class Dataset:
    x: pd.DataFrame
    y: pd.DataFrame
    name: str = "unknown"
    meta_data: Optional[dict] = None

    def __repr__(self):
        return f' Dataset {self.name}  x.shape = {self.x.shape}  y.shape = {self.y.shape}  meta_data = {self.meta_data}'

    def __getitem__(self, is_in: pd.Series):
        """ Subset of the dataset

        Parameters
        ----------
        is_in: pd.Series
            Pandas Series with bool values indicating if a row should be
            selected.

        Returns
        -------
        dataset: Dataset
            New Dataset where x and y are chosen using `is_in`. Other
            attributes are the same.

        Notes
        -----
        A deep copy is not performed.
        """
        is_in = is_in.reindex_like(self.x)
        if is_in.hasnans:
            warn('It seems that `is_in` index is not consistent with dataset '
                 'index. Rows that do not have corresponding values in `is_in` '
                 'are dropped.')
            is_in.fillna(False, inplace=True)
        return Dataset(x=self.x[is_in], y=self.y[is_in],
                       name=self.name, meta_data=self.meta_data)

    @staticmethod
    def from_target_columns(df: pd.DataFrame, target_columns: List[str],
                            name: str = "unknown",
                            meta_data: Optional[dict] = None,
                            allow_empty_y: bool = False) -> Dataset:
        """ Factory method for creating a Dataset from list of targets.

        Parameters
        ----------
        df: pd.DataFrame
            DataFrame that contains all the data: features and targets
        target_columns: list of strings
            Columns that should become targets
        name: str, optional
            Name of the dataset
        meta_data: dict, optional
            Optional metadata
        allow_empty_y: bool, optional
            If True, will not raise an error when names from target column
            lists are not in df or if the `target_columns` is an empty list.
            In such cases, it will create a dataset with empty targets and same
            features as in df.

        Returns
        -------
        dataset: Dataset
            An object with features and targets separated.
        """
        data = df.copy()

        try:
            y = pd.concat([data.pop(c) for c in target_columns], 1)
        except (KeyError, ValueError):
            if allow_empty_y:
                return Dataset(x=df.copy(), y=None, name=name, meta_data=meta_data)
            raise

        return Dataset(x=data, y=y, name=name, meta_data=meta_data)



# Copyright 2020 (c) Netguru S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import contextlib
import os
from abc import ABC, abstractmethod
from collections import namedtuple
from typing import Union, Callable, List, Dict, Tuple, Optional

import mlflow
import numpy as np
import optuna
from sklearn.metrics import SCORERS
from sklearn.model_selection import StratifiedKFold, StratifiedShuffleSplit

from opoca.data.dataset import Dataset
from opoca.hyperparam_search.utils import check_mlflow, get_current_git_hash, \
    are_uncommitted_changes, are_untracked_files, mlflow_push_pickled, \
    MLFLOW_TRACKING_URI

NamedScorer = namedtuple('NamedScorer', ['fn', 'name'])


def _check_scorer(scorer: Union[str, Callable]) -> NamedScorer:
    """ Check and standardize a scorer """
    if isinstance(scorer, str):
        scorer_fn = SCORERS[scorer]
        scorer_name = scorer
    elif isinstance(scorer, Callable):
        scorer_fn = scorer
        scorer_name = scorer.__name__
    else:
        raise ValueError('Scorer must be either a str or a callable.')

    return NamedScorer(scorer_fn, scorer_name)


class Objective(ABC):
    """ Callable base class for implementing objective function.

    You should subclass this class and implement `create_model`. It was
    designed to provide decent hyperparameter optimization out of the box.

    Parameters
    ----------
    dataset: Dataset
        POC-native dataset with training data.
    num_cv_folds: int, optional
        Number of folds to use in CV. If set, `val_split_ratio` must be None.
    val_split_ratio: float, optional
        Percentage of the data to be used for validation. If set,
        `num_cv_folds` must be None.
    scorer: str or sklearn scorer or list of those
        Check possible scores in  sklearn.metrics.SCORERS.
        If list, the first scorer will be used as the optimization metric, the
        rest just for monitoring.
    unique_name: str, optional
        Name used to distinguish consecutive experiments in mlflow and studies
        in optuna. If you use existing name, optuna will continue study (if it
        was recorded in database) and mlflow will log under the same name.
        If None, the name will be generated based on the objective name and git
        status, see the `_get_name` doc for details.
    params_to_mlflow: bool, optional
        If True,  parameters will be logged in mlflow server.
    models_to_mlflow: bool, optional
        If True, pickled models will be pushed in mlflow server.
    fit_kwargs: dict, optional
        Kwargs that will be passed to fit method.

    If neither num_cv_folds nor val_split_ratio is set, a default
    value num_cv_folds=5 will be used.

    If any of mlflow flags are True, you must specify following variables in
    your environment:
    - MLFLOW_TRACKING_URI
    - MLFLOW_TRACKING_USERNAME
    - MLFLOW_TRACKING_PASSWORD
    """

    def __init__(self, dataset: Dataset,
                 num_cv_folds: int = None,
                 val_split_ratio: float = None,
                 scorer: Union[str, Callable, list] = 'f1_weighted',
                 unique_name: Optional[str] = None,
                 params_to_mlflow: bool = False,
                 models_to_mlflow: bool = False,
                 fit_kwargs: Optional[dict] = None):

        if num_cv_folds is not None and val_split_ratio is not None:
            raise TypeError('Both `num_cv_folds` and `val_split_ratio` were '
                            'given. Specify at most one of them.')

        if val_split_ratio is not None:
            sss = StratifiedShuffleSplit(n_splits=1, test_size=val_split_ratio)
            self.split_indices = [next(sss.split(dataset.x, dataset.y))]
        else:
            num_cv_folds = num_cv_folds or 5
            cross_validation = StratifiedKFold(n_splits=num_cv_folds, shuffle=True)
            self.split_indices = list(cross_validation.split(dataset.x, dataset.y))

        self.dataset = dataset

        if isinstance(scorer, list):
            scorer, *other_scorers = scorer
        else:
            other_scorers = []
        self.scorer, self.scorer_name = _check_scorer(scorer)
        self.additional_named_scorers = [_check_scorer(s) for s in other_scorers]
        if len(self.additional_named_scorers) > 0:
            print(f'Only {self.scorer_name} will be used for optimization. '
                  f'Other scores will be calculated and logged, but will not '
                  f'influence the hyperparameter search')

        self.name = self._get_name(unique_name)

        self.params_to_mlflow = params_to_mlflow
        self.models_to_mlflow = models_to_mlflow
        if self.use_mlflow:
            if not check_mlflow():
                raise ValueError('MLflow environment variables not set, cannot use it.')
            print(f'Trials will be logged to MLflow at {os.environ[MLFLOW_TRACKING_URI]}.')
            mlflow.set_experiment(self.name)

        self.fit_kwargs = fit_kwargs if fit_kwargs is not None else {}

    @property
    def use_mlflow(self) -> bool:
        """ Property indicating if mlflow is used at all """
        return any((self.params_to_mlflow, self.models_to_mlflow))

    def _get_name(self, unique_name):
        """ Assure decent name for an experiment (aka study)

        It generates name of the study based on:
        - the name of objective class (one that inherits from this one)
        - current git hash (7 characters)
        - git status by adding:
            - 'dirty' if there are some uncommitted changes,
            - 'with_untracked' if there are some untracked files.

        In other words, if no suffix is present, you can be sure that the code
        used to train model is exactly as in the commit.

        Parameters
        ----------
        unique_name: str or None
            If provided, it is returned immediately.
            If None, it is generated as described above.

        Returns
        -------
        name: str
            Provided of generated name.
        """
        if unique_name is not None:
            return unique_name
        git_hash = get_current_git_hash()
        if git_hash is None:
            return f'{self.__class__.__name__}/not-in-git'
        name = f"{self.__class__.__name__}/{git_hash}" \
               f"{'_dirty' if are_uncommitted_changes() else ''}" \
               f"{'_with_untracked' if are_untracked_files() else ''}"

        return name

    def __call__(self, trial: optuna.Trial) -> float:
        """ Make the class Callable as required by optuna.Study

        It handles most of mlflow reporting.

        Parameters
        ----------
        trial: optuna.Trial
            Object that controls how parameters are drawn.

        Returns
        -------
        score: float
            Score, averaged over folds when applicable.
        """
        if self.use_mlflow:
            context = mlflow.start_run(run_name=f'trial-{trial.number:03d}')
        else:
            context = contextlib.nullcontext()

        with context:
            model = self.create_model(trial)

            if self.params_to_mlflow:
                mlflow.log_params(trial.params)
                mlflow.log_param('dataset', self.dataset.name)

            scores, additional_scores = self._cross_validate(model, trial)
            mean_score = float(np.mean(scores))
            mean_additional_scores = {f'mean_{name}': float(np.mean(s))
                                      for name, s in additional_scores.items()}

            if self.use_mlflow and len(scores) > 1:
                mlflow.log_metrics({'MEAN_' + self.scorer_name: mean_score})
                mlflow.log_metrics(mean_additional_scores)
            if len(mean_additional_scores) > 0:
                print(*[f'{name}: {v:.3g}  ' for name, v in mean_additional_scores.items()])

            return mean_score

    def _cross_validate(self, model, trial: optuna.Trial) \
            -> Tuple[List[float], Dict[str, List[float]]]:
        """ Validate model using validation scheme defined in `__init__`.

        It handles pruning and some mlflow reporting.

        Parameters
        ----------
        model: sklearn estimator
            Model returned by `create_model` specific for the subclass.
        trial: optuna.Trial
            Object that controls how parameters are drawn.

        Returns
        -------
        List of scores and dict of additional scores.

        Raises
        ------
        optuna.exceptions.TrialPruned
            when trial is not promising and should be pruned. The exception is
            further handled by optuna.
        """
        x, y = self.dataset.x, self.dataset.y
        scores = []
        additional_scores = {ans.name: [] for ans in self.additional_named_scorers}
        for i, (train_index, val_index) in enumerate(self.split_indices, start=1):
            x_train, x_val = x.iloc[train_index], x.iloc[val_index]
            y_train, y_val = y.iloc[train_index], y.iloc[val_index]

            model.fit(x_train, y_train, **self.fit_kwargs)

            score = self.scorer(model, x_val, y_val)
            scores.append(score)
            if self.use_mlflow:
                mlflow.log_metrics({self.scorer_name: score}, step=i)

            for fn, name in self.additional_named_scorers:
                additional_score = fn(model, x_val, y_val)
                additional_scores[name].append(additional_score)
                if self.use_mlflow:
                    mlflow.log_metrics({name: additional_score}, step=i)

            if self.models_to_mlflow:
                mlflow_push_pickled(model, f'model_cv{i:02d}_')

            trial.report(score, i)
            if trial.should_prune():
                if self.use_mlflow:
                    mlflow.set_tag('PRUNED_FOLD', str(i + 1))
                raise optuna.exceptions.TrialPruned()

        return scores, additional_scores

    @abstractmethod
    def create_model(self, trial: optuna.Trial):
        """ Return a model with parameters drawn using trial.

        This method should be implemented in the subclass. See `examples.py`.

        Parameters
        ----------
        trial: optuna.Trial
            Object that should be used in the implementation to get parameters
            using `suggest_*` methods, e.g.:
            ```
            C = trial.suggest_loguniform('C', 1e-6, 1e6)
            kernel = trial.suggest_categorical('kernel', ['linear', 'poly', 'rbf', 'sigmoid'])
            model = SVC(C=C, kernel=kernel)
            ```

        Returns
        -------
        model: estimator
        """



# Copyright 2020 (c) Netguru S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import warnings
from typing import List, Tuple

import numpy as np
from sklearn.metrics import roc_curve, auc, precision_recall_curve, average_precision_score
from sklearn.preprocessing import OneHotEncoder


def convert_list_probas_to_array(probabilities_list: list) -> np.ndarray:
    """
    Converts list of probability arrays to a single array of probabilities.

    In multilabel tasks the model's predict_proba method might output the probabilities of a class being
    a 0 or a 1 as separate arrays - the output is list of length of number of classes.
    The elements of the list are 2 dimensional arrays of probabilities.

    Parameters
    ----------
    probabilities_list: List of numpy ndarrays
        List of length of number of classes. The elements of the list are 2 dimensional arrays of probabilities.

    Returns
    -------
    out: np.ndarray
        Single array of probabilities.
        The last shape of the output array is equal to number of classes.
    """
    assert isinstance(probabilities_list, list)

    probabilities = np.empty(shape=(probabilities_list[0].shape[0], len(probabilities_list)))
    for i, array in enumerate(probabilities_list):
        if isinstance(array, np.ndarray):
            probabilities[:, i] = array[:, 1]
    return probabilities


def calculate_roc_auc(y_true: np.ndarray, y_prob: np.ndarray, class_names: List[str],
                      task: str = None) -> Tuple[dict, dict, dict]:
    """
    Calculates Receiver Operating Characteristic Area Under Curve.
    Apart from ROC scores the function outputs the False Positives Ratios,
    and True Positive Ratios, which can be used to plot ROC curves.
    Parameters
    ----------
    y_true: np.ndarray
        Array of true labels (last shape of array should equal to number of classes).
    y_prob: np.ndarray
        Array of predicted probabilities (last shape of array should equal to number of classes).
    class_names: List of strings
        Names of classes.
    task: str, optional
        Type of classification task. Types of task supported: ['binary', 'multiclass', 'multilabel'].
    Returns
    -------
    out: Tuple of 3 dictionaries
        The outputs are 3 dictionaries: False Positives Ratios and True Positive Ratios which are dictionaries
        in which keys are class names and values are numpy arrays, and a dictionary of ROC AUC scores.
    """
    fpr = dict()
    tpr = dict()
    roc_auc_dict = dict()

    if not task:
        task = check_task(y_true, y_prob)

    if task in ['binary', 'multiclass']:
        encoder = OneHotEncoder()
        y_true = encoder.fit_transform(y_true).toarray()

    for i, name in enumerate(class_names):
        fpr[name], tpr[name], _ = roc_curve(y_true[:, i], y_prob[:, i])
        roc_auc_dict[name] = auc(fpr[name], tpr[name])

    # Compute micro-average ROC curve and ROC area
    fpr["micro avg"], tpr["micro avg"], _ = roc_curve(y_true.ravel(), y_prob.ravel())
    if task == 'binary':
        # if task is binary avg roc auc is the same as roc auc of both classes
        # therefore we use the calculated value of one of the classes to fill it
        roc_auc_dict["micro avg"] = roc_auc_dict[class_names[0]]
    else:
        roc_auc_dict["micro avg"] = auc(fpr["micro avg"], tpr["micro avg"])

    fpr, tpr, roc_auc_dict = calculate_macro_auc(fpr, tpr, roc_auc_dict, class_names)

    return fpr, tpr, roc_auc_dict


def calculate_macro_auc(fpr: dict, tpr: dict, roc_auc_dict: dict, class_names: List[str]):
    """
    Calculates Receiver Operating Characteristic Area Under Curve macro average.
    Parameters
    ----------
    fpr: dict
        Dictionary in which keys are names of classes and values are False Positives Ratio arrays.
    tpr: dict
        Dictionary in which keys are names of classes and values are True Positives Ratio arrays.
    roc_auc_dict: dict
        Dictionary in which keys are names of classes and values are calculated ROC area under curve.
    class_names: List of strings
        Names of classes.

    Returns
    -------
    out: Tuple of 3 dictionaries
        The outputs are 3 dictionaries: False Positives Ratios and True Positive Ratios which are dictionaries
        in which keys are names and values are numpy arrays, and a dictionary of ROC AUC scores with 'macro avg' added.
    """
    nans_list = [name for name in class_names if any(np.isnan(fpr[name])) or any(np.isnan(tpr[name]))]

    # If some classes contain NaNs remove them from further calculations.
    if nans_list:
        warnings.warn(f'Classes {nans_list} contain NaN values. Ignoring them in macro average calculations.')

    # First aggregate all false positive rates
    all_fpr = np.unique(np.concatenate([fpr[name] for name in class_names if name not in nans_list]))

    # Then interpolate all ROC curves at this points
    mean_tpr = np.zeros_like(all_fpr)
    for name in class_names:
        if name not in nans_list:
            mean_tpr += np.interp(all_fpr, fpr[name], tpr[name])

    # Finally average it and compute AUC
    mean_tpr /= (len(class_names) - len(nans_list))

    fpr["macro avg"] = all_fpr
    tpr["macro avg"] = mean_tpr
    roc_auc_dict["macro avg"] = auc(fpr["macro avg"], tpr["macro avg"])

    return fpr, tpr, roc_auc_dict


def check_task(y_true: np.ndarray, y_prob: np.ndarray) -> str:
    """
    Checks classification task type based on shapes of true and predicted probabilities arrays.
    Parameters
    ----------
    y_true: np.ndarray
        Array of true labels (last shape of array should equal to number of classes).
    y_prob: np.ndarray
        Array of predicted probabilities (last shape of array should equal to number of classes).

    Returns
    -------
    out: str
        Type of task as string. Possible outputs: ['binary', 'multiclass', 'multilabel'].
    """
    if y_prob.shape[-1] == 2 and len(np.unique(y_true)) == 2:
        task = 'binary'
    elif len(np.unique(y_true)) > 2 and y_prob.shape[-1] > 2:
        task = 'multiclass'
    else:
        task = 'multilabel'
    return task


def calculate_prauc(y_true: np.ndarray, y_prob: np.ndarray) -> Tuple[dict, dict, dict]:
    """
    Calculates Precision-Recall Area Under Curve.
    Parameters
    ----------
    y_true: np.ndarray
        Array of true labels (last shape of array should equal to number of classes).
    y_prob: np.ndarray
        Array of predicted probabilities (last shape of array should equal to number of classes).

    Returns
    -------
    out: Tuple of 3 dictionaries
        The outputs are 3 dictionaries. Precision and Recall curves and Average Precision scores.
    """
    precision = dict()
    recall = dict()
    average_precision = dict()

    task = check_task(y_true, y_prob)

    if task in ['binary', 'multiclass']:
        encoder = OneHotEncoder()
        y_true = encoder.fit_transform(y_true).toarray()

    for i in range(y_prob.shape[-1]):
        precision[i], recall[i], _ = precision_recall_curve(y_true[:, i], y_prob[:, i])
        average_precision[i] = average_precision_score(y_true[:, i], y_prob[:, i])

    # A "micro-average": quantifying score on all classes jointly
    precision["micro"], recall["micro"], _ = precision_recall_curve(y_true.ravel(), y_prob.ravel())
    average_precision["micro"] = average_precision_score(y_true, y_prob, average="micro")

    return precision, recall, average_precision



# Copyright 2020 (c) Netguru S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module collects common targets transformers
"""
from typing import Union

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_is_fitted


class ToNumpy(TransformerMixin, BaseEstimator):
    """
    Converts pd.DataFrame targets to numpy array
    """
    def fit(self, y: Union[pd.DataFrame, pd.Series]):
        if isinstance(y, pd.DataFrame):
            self.columns_ = y.columns
        else:
            self.columns_ = [y.name]
        self.index_ = y.index
        return self

    def transform(self, y: pd.DataFrame) -> np.ndarray:
        check_is_fitted(self)
        return y.to_numpy()

    def inverse_transform(self, y: np.ndarray) -> pd.DataFrame:
        return pd.DataFrame(y, columns=self.columns_, index=self.index_)



# Copyright 2020 (c) Netguru S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import List, Dict, Any

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_is_fitted
import pandas as pd


def get_col_stat(df: pd.DataFrame, column: str, stat: str) -> float:
    """
    Calculates statistics of a given column from given DataFrame.

    Parameters
    ----------
    df : pandas.DataFrame
        DataFrame containing given column.
    column : str
        Column name given as string.
    stat: str
        Name of the statistic to calculate.
        Available statistics are: ['min', 'max', 'mean', 'mode', 'median']

    Returns
    -------
    Depending of the data type in the column the result might be a float, int, or a string.

    """
    assert stat in ['min', 'max', 'mean', 'mode', 'median'], 'The stat is not supported'
    if stat == 'mode':
        result = df[column].value_counts().idxmax()
    else:
        result = df[column].apply(stat)
    return result


class MissingDataHandler(BaseEstimator, TransformerMixin):
    """
    Stores names of columns with missing data in objective specific lists or dictionaries.
    """

    def __init__(self, to_drop: List[str] = None, to_fill_with_value: Dict[str, Any] = None,
                 to_fill_with_stat: Dict[str, List[str]] = None):
        """
        Parameters
        ----------
        to_drop: list
            List of column names to drop from DataFrame
        to_fill_with_value: dict
            Dictionary whose keys are column names, and values are values to fill missing data with in that column.
            Examples
            >>> to_fill_with_value={'income': 11000, 'sex': 'F'}
        to_fill_with_stat: dict
            Dictionary whose keys are stat to calculate, and values are lists of columns (str) to fill with that stat.
            Supported stats are: ['min', 'max', 'mean', 'mode', 'median']
            Examples
            >>> to_fill_with_stat={'mean': ['revenue', 'income'], 'median': ['age']}
        """

        self.to_drop = to_drop if to_drop else []
        self.to_fill_with_value = to_fill_with_value if to_fill_with_value else {}
        self.to_fill_with_stat = to_fill_with_stat if to_fill_with_stat else {}

    def fit(self, df: pd.DataFrame):
        for stat, columns in self.to_fill_with_stat.items():
            for column in columns:
                calculated = get_col_stat(df=df, column=column, stat=stat)
                self.to_fill_with_value[column] = calculated
        self.fitted_ = True  # fitted_ flag - for check_is_fitted function in transform method
        return self

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        check_is_fitted(self)
        df.drop(self.to_drop, axis=1, inplace=True)
        for column in set(df.columns).intersection(self.to_fill_with_value):
            df[column].fillna(self.to_fill_with_value[column], inplace=True)
        return df



# Copyright 2020 (c) Netguru S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import pickle
import subprocess
import warnings
from tempfile import NamedTemporaryFile
from typing import Optional

import mlflow

OPTUNA_DB = 'OPTUNA_DB'
MLFLOW_TRACKING_URI = 'MLFLOW_TRACKING_URI'
MLFLOW_TRACKING_USERNAME = 'MLFLOW_TRACKING_USERNAME'
MLFLOW_TRACKING_PASSWORD = 'MLFLOW_TRACKING_PASSWORD'


def check_mlflow() -> bool:
    """ Check if mlflow variables are set withing the current environment.

    It checks following variables:
    - MLFLOW_TRACKING_URI
    - MLFLOW_TRACKING_USERNAME
    - MLFLOW_TRACKING_PASSWORD

    Returns
    -------
    True, when all variables are set.
    False, otherwise

    Notes
    -----
    It does not check if the connection is actually working.

    """
    try:
        _ = os.environ[MLFLOW_TRACKING_URI]
        _ = os.environ[MLFLOW_TRACKING_USERNAME]
        _ = os.environ[MLFLOW_TRACKING_PASSWORD]
    except KeyError:
        return False

    return True


def mlflow_push_pickled(obj, name: str = '', put_in_dir: Optional[str] = None):
    """ Push a picklable object to s3 storage associated with mlflow

    It should be used within a mlflow run context.

    Parameters
    ----------
    obj:
        Any picklable object.
    name: str, optional
        File name will be constructed as: this name + 7 random chars + '.pickle'
    put_in_dir: str that is a correct directory name, optional
        If provided, the object will be placed in a folder with this name.
    """
    with NamedTemporaryFile(prefix=name, suffix='.pickle') as tempf:
        pickle.dump(obj, tempf)
        mlflow.log_artifact(tempf.name, artifact_path=put_in_dir)


def get_current_git_hash(raise_on_error: bool = False) -> Optional[str]:
    """ Return git hash of the latest commit

    Parameters
    ----------
    raise_on_error: bool, optional
        If False (default), will return None, when it fails to obtain commit hash.
        If True, will raise, when it fails to obtain commit hash.

    Returns
    -------
    Short hash of the current HEAD or None.
    """
    try:
        git_hash = subprocess.check_output(['git', 'rev-parse', '--short', 'HEAD']).decode('utf-8').strip()
    except subprocess.CalledProcessError:
        if raise_on_error:
            raise
        warnings.warn('Probably not in a git repo.')
        git_hash = None

    return git_hash


def are_uncommitted_changes() -> bool:
    """ Check if there are any uncommitted changes

    Returns
    -------
    Bool indicating if there are any changes. False means no changes.

    Notes
    -----
    It does not look at untracked files.
    `git update-index --refresh` makes touched but not changed files to be treated correctly (as unchanged),
        ref.: https://stackoverflow.com/a/3879077/8788960
    """
    subprocess.call(['git', 'update-index', '--refresh'], stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)
    changes = subprocess.call(['git', 'diff-index', '--quiet', 'HEAD', '--'])

    return bool(changes)


def are_untracked_files() -> bool:
    """ Check if there are any untracked files

    Returns
    -------
    Bool indicating if there are any untracked files. False means no files.
    """
    untracked = subprocess.check_output(['git', 'ls-files', '--others', '--exclude-standard']).decode('utf-8').strip()

    return bool(untracked)



# Copyright 2020 (c) Netguru S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Sklearn flavor module that collects trainers implementations that supports sklearn-like models
"""
import inspect
import warnings
from abc import ABC
from typing import Dict, Callable, Union

from imblearn.base import BaseSampler
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.pipeline import make_pipeline, Pipeline

from opoca.data.targets import ToNumpy


class BaseTrainer(ABC):
    """
    Abstract class for sklearn-like trainer
    """

    def __init__(self, features_pipeline: Union[BaseEstimator, TransformerMixin],
                 clf: BaseEstimator):
        self.features_pipeline = features_pipeline
        self.clf = clf


class SupervisedTrainer(BaseTrainer):
    """
    Class that manages training models in supervised manner

    Parameters
    ----------
    features_pipeline: [BaseEstimator, TransformerMixin]
        Pipeline for transforming input data to features

    clf: BaseEstimator
        Classifier sklearn-like

    targets_transformer: default acts like identity function
        Transforms targets vector if needed (for example to one-hot encoding)

    sampler: BaseSampler, default returns unchanged dataset
        Resamples data, supports various oversampling and undersampling techniques from imblearn

    clf_fit_params: Dict, optional
        Dictionary with keywords args for classifier fit method

    features_pipeline_callback: Callable, optional
        Might be used for updating pipeline between fit and transform invokes
    """

    def __init__(self, features_pipeline: Union[BaseEstimator, TransformerMixin, Pipeline],
                 clf: BaseEstimator,
                 targets_transformer: Union[TransformerMixin, BaseEstimator] = None,
                 sampler: BaseSampler = None,
                 clf_fit_params: Dict = None,
                 features_pipeline_callback: Callable = None
                 ):
        super().__init__(features_pipeline, clf)

        self.features_pipeline_callback = features_pipeline_callback

        if clf_fit_params is None:
            clf_fit_params = {}
        self.clf_fit_params = clf_fit_params

        if targets_transformer is None:
            targets_transformer = ToNumpy()
        self.targets_transformer = targets_transformer

        self.sampler = sampler

    def train(self, train_dataset, val_dataset=None, fit_data=None):
        """
        Runs complete training. Currently val_dataset is used only in case of early stopping for XGBoost.

        Parameters
        ----------
        train_dataset: Dataset
            Instance of Dataset class that represents training set

        val_dataset: Dataset, optional
            Instance of Dataset class that represents validation set

        fit_data: pd.DataFrame, optional
            DataFrame that is to be used for fitting features pipeline

        Returns
        -------
        pipeline: Pipeline
            Complete fitted pipeline
        """

        fit_data = fit_data if fit_data is not None else train_dataset.x

        # first we fit features pipeline
        self.features_pipeline.fit(fit_data)

        self.targets_transformer.fit(train_dataset.y)

        # if needed, we update features pipeline (typical case when you fit features on test time-series-data
        if self.features_pipeline_callback is not None:
            self.features_pipeline_callback(self.features_pipeline)

        if self.sampler is not None:
            x_train, y_train = self.sampler.fit_resample(train_dataset.x, train_dataset.y)
        else:
            x_train, y_train = train_dataset.x, train_dataset.y

        x_features = self.features_pipeline.transform(x_train)
        y_targets = self.targets_transformer.transform(y_train)

        if val_dataset is not None:
            self.__prepare_early_stop_params(val_dataset)

        self.clf.fit(x_features, y_targets, **self.clf_fit_params)

        pipeline = make_pipeline(self.features_pipeline,
                                 self.clf)

        return pipeline

    def __prepare_early_stop_params(self, val_dataset):
        """

        Parameters
        ----------
        val_dataset: Dataset
            Validation dataset used for early stopping
        """

        if "eval_set" in inspect.signature(self.clf.fit).parameters:
            val_y = self.targets_transformer.fit_transform(val_dataset.y)
            val_x = self.features_pipeline.transform(val_dataset.x)

            self.clf_fit_params["eval_set"] = [(val_x, val_y)]

            self.clf_fit_params["eval_metric"] = self.clf_fit_params["eval_metric"] or ["logloss"]

            self.clf_fit_params["early_stopping_rounds"] = self.clf_fit_params["early_stopping_rounds"] or 10
        else:
            warnings.warn(f"There's no 'eval_set' parameter in fit method of instance "
                          f"of {self.clf.__class__}. Providing validation set will have no effect on "
                          f"training", UserWarning)



# Copyright 2020 (c) Netguru S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import warnings

import numpy as np
from joblib import Parallel, delayed
from sklearn.multiclass import OneVsRestClassifier, _ConstantPredictor, clone
from sklearn.preprocessing import LabelBinarizer


def _fit_binary(estimator, x, y, classes=None, sample_weight=None, eval_set=None, eval_metric=None,
                early_stopping_rounds=None, verbose=True, xgb_model=None, sample_weight_eval_set=None, callbacks=None):
    """Fit a single binary estimator.
    This function is ported from sklearn multiclass module."""
    unique_y = np.unique(y)
    if len(unique_y) == 1:
        if classes is not None:
            if y[0] == -1:
                c = 0
            else:
                c = y[0]
            warnings.warn(f"Label {str(classes[c])} is present in all training examples.")
        estimator = _ConstantPredictor().fit(x, unique_y)
    else:
        estimator = clone(estimator)
        estimator.fit(x, y, sample_weight=sample_weight, eval_set=eval_set, eval_metric=eval_metric,
                      early_stopping_rounds=early_stopping_rounds, verbose=verbose,
                      xgb_model=xgb_model, sample_weight_eval_set=sample_weight_eval_set, callbacks=callbacks)
    return estimator


class XGBoostOneVsRest(OneVsRestClassifier):

    def fit(self, X, y, sample_weight=None, eval_set=None, eval_metric=None, early_stopping_rounds=None,
            verbose=True, xgb_model=None, sample_weight_eval_set=None, callbacks=None):
        """Fit underlying estimators.
        Parameters
        ----------
        X : (sparse) array-like of shape (n_samples, n_features)
            Data.
        y : (sparse) array-like of shape (n_samples,) or (n_samples, n_classes)
            Multi-class targets. An indicator matrix turns on multilabel
            classification.
        Other Parameters: see XGBoostClassifer documentation.
        Returns
        -------
        self
        """
        # A sparse LabelBinarizer, with sparse_output=True, has been shown to
        # outperform or match a dense label binarizer in all cases and has also
        # resulted in less or equal memory consumption in overall
        self.label_binarizer_ = LabelBinarizer(sparse_output=True)
        Y = self.label_binarizer_.fit_transform(y)
        Y = Y.tocsc()
        self.classes_ = self.label_binarizer_.classes_
        columns = (col.toarray().ravel() for col in Y.T)
        # In cases where individual estimators are very fast to train setting
        # n_jobs > 1 in can results in slower performance due to the overhead
        # of spawning threads.  See joblib issue #112.
        jobs = []
        for i, column in enumerate(columns):
            class_name = self.label_binarizer_.classes_[i]
            classes = [f"not {class_name}", class_name]
            job_eval_set = [(x_set, y_set[:, i]) for (x_set, y_set) in eval_set]

            delayed_job = delayed(_fit_binary)(self.estimator, X, column, classes, sample_weight, job_eval_set,
                                               eval_metric, early_stopping_rounds, verbose, xgb_model,
                                               sample_weight_eval_set, callbacks)
            jobs.append(delayed_job)

        self.estimators_ = Parallel(n_jobs=self.n_jobs)(jobs)
        return self



# Copyright 2020 (c) Netguru S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module that collects functions and classes for inputs and targets preprocessing.

Available classes:
* Dataset
* Split
* Preprocessor
* TargetTransformer
* IdentityTransformer

"""
from __future__ import annotations

from abc import abstractmethod, ABC
from collections import defaultdict
from typing import Callable, Tuple, List
from typing import Dict, Union

import numpy as np
import pandas as pd
import scipy.stats as ss
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import cross_val_score
from sklearn.preprocessing import OneHotEncoder

from opoca.features.transformers import NaNColumnDropper, SingleValueColumnDropper, NanReplacer


def find_correlated_with_target(dataset: pd.DataFrame, target_name: str = "Target", threshold: float = 0.8,
                                cv: int = 3, scoring: Union[Callable, str] = 'f1') -> List[Tuple[str, float]]:
    """
    Finds variables that have `scoring` (default to f1-score) higher than `threshold` when logistic regression is
    trained with them as single input variable.

    Parameters
    ----------
    dataset:
        Input dataset
    target_name:
        Name of column that contain target variable
    threshold:
        Determines if variables is too much correlated with target or not
    cv:
        Number of folds in cross-validation
    scoring:
        Metric to determine whether variable is correlated with target or not

    Returns
    -------
    List of correlated variables along with scores
    """
    correlated_columns = []
    y = dataset[target_name]

    for col in dataset.columns:
        if col == target_name:
            continue
        if dataset[col].dtype == "object":
            one_hot_encoder = OneHotEncoder()
            x = one_hot_encoder.fit_transform(dataset[col])
        else:
            x = dataset[col].values.reshape(-1, 1)
        clf = LogisticRegression()
        score = cross_val_score(clf, x, y, cv=cv, scoring=scoring).mean()
        if score > threshold:
            correlated_columns.append((col, score))

    return correlated_columns


def find_correlated_variables(data_frame: pd.DataFrame, kind: str = "categorical", threshold: float = 0.9) -> Dict:
    """
    Computes which variables are correlated more than threshold. Works for categorical-categorical and
    continuous-continuous variables.

    Parameters
    ----------
    data_frame:
        input data frame
    kind:
        categorical or continuous
    threshold:
        Returns variables that have correlation coefficient higher or equal to threshold

    Returns
    -------
    Dictionary where key is variable name and value is list of variables correlated with key variable more than
    threshold.
    """
    correlated_var2var_dict = defaultdict(list)

    def pearson_coeff(x, y):
        result, _ = ss.pearsonr(x, y)
        return result

    if kind == "categorical":
        compute_correlation_coeff = cramers_corrected_stat
        types = ["object", "b"]
    elif kind == "continuous":
        types = ["float", "int64"]
        compute_correlation_coeff = pearson_coeff
    else:
        raise ValueError("Incorrect kind of variable")

    df = data_frame.select_dtypes(types)

    for i in range(len(df.columns)):
        for j in range(i + 1, len(df.columns)):
            col_1 = df.columns[i]
            col_2 = df.columns[j]
            try:
                res = compute_correlation_coeff(df[col_1], df[col_2])
                if abs(res) >= threshold:
                    correlated_var2var_dict[col_1].append(col_2)
                    correlated_var2var_dict[col_2].append(col_1)
            except ValueError:
                continue

    return correlated_var2var_dict


def cramers_corrected_stat(x: Union[np.ndarray, pd.Series], y: Union[np.ndarray, pd.Series]) -> float:
    """
    calculate Cramers V statistic for categorial-categorial association.
    uses correction from Bergsma and Wicher,
    Journal of the Korean Statistical Society 42 (2013): 323-328
    """
    eps = 1.e-6
    confusion_matrix = pd.crosstab(x, y)
    if confusion_matrix.shape[0] == 2:
        correct = False
    else:
        correct = True
    chi2 = ss.chi2_contingency(confusion_matrix, correction=correct)[0]

    n = sum(confusion_matrix.sum())
    phi2 = chi2 / n
    r, k = confusion_matrix.shape
    phi2corr = max(0, phi2 - ((k - 1) * (r - 1)) / (n - 1 + eps))
    rcorr = r - ((r - 1) ** 2) / (n - 1 + eps)
    kcorr = k - ((k - 1) ** 2) / (n - 1 + eps)
    result = np.sqrt(phi2corr / (eps + min((kcorr - 1), (rcorr - 1))))
    return result


class Preprocessor(TransformerMixin, BaseEstimator, ABC):
    """
    Abstract class for all preprocessors
    """

    @abstractmethod
    def fit(self, df):
        pass

    @abstractmethod
    def transform(self, df):
        pass


class BasicPreprocessor(Preprocessor):
    """
    Basic preprocessor that includes:
    * dropping predefined columns
    * dropping columns with too many NaNs
    * replacing various representations of NaNs with np.nan
    * dropping single value columns

    Parameters
    ----------
    rules:
        dictionary of preprocessing rules. Under key `DROP` there is list of columns that shall be dropped, under key
        `STAY` there is list of columns that shall stay regardless of other partial preprocessors
    drop_nan_threshold
        drop columns that have more NaNs fraction that this threshold
    to_replace
        list of NaNs representations, if None then take a look into `NanReplacer` docs to find default representations
    """

    def __init__(self, rules: Dict, drop_nan_threshold: float = 0.8, to_replace: List[str] = None):
        self.rules = rules
        self.ncd = NaNColumnDropper(threshold=drop_nan_threshold)
        self.svcd = SingleValueColumnDropper()
        self.replace_nan = NanReplacer(to_replace=to_replace)
        self.drop_columns_ = None

    def fit(self, df: pd.DataFrame) -> BasicPreprocessor:
        self.ncd.fit(df)
        self.drop_columns_ = set(self.rules["DROP"]).union(set(self.ncd.columns_to_drop_)) - set(self.rules["STAY"])
        return self

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        df.drop(columns=self.drop_columns_, inplace=True)

        df = self.replace_nan.fit_transform(df)
        df = self.svcd.fit_transform(df)

        return df



# Copyright 2020 (c) Netguru S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations  # for correct annotation of forward reference in Split

from dataclasses import dataclass
from enum import IntEnum

import pandas as pd

from opoca.data.dataset import Dataset

SPLIT_FLAG_NAME = 'split_flag'


class SplitFlag(IntEnum):
    TRAIN = 0
    VAL = 1
    TEST = 2
    # IGNORE  can be used to mark rows that should be  omitted completely,
    # e.g. with wrong labels
    IGNORE = -1


@dataclass
class Split:
    """ A simple dataclass used for handling train/val/test splits

    `to_split_table` allows exporting the current split to a human-readable
    format and further saving it as csv. If a file name corresponds to data set
    file name, the split can be automatically reloaded with `DataHandler.load`
    method with `return_as='split'`.
    """
    train: Dataset
    val: Dataset
    test: Dataset

    def __iter__(self):
        return iter([self.train, self.val, self.test])

    def __str__(self):
        return f'Split\n  * train: {self.train}\n  * val:   {self.val}\n  ' \
               f'* test:  {self.test}\n'

    @classmethod
    def from_split_table(cls, dataset: Dataset, split_table: pd.Series) -> Split:
        """ Factory method to create a Split object from split table

        Parameters
        ----------
        dataset: Dataset
            Dataset that contains all the data or its subsample.
        split_table: pd.Series
            Series with int values indicating to which dataset a given row
            belongs. See `SplitFlag` enum for details. Any value that is not in
            SplitFlag is ignored.

        Notes
        -----
        Every element of the index in `dataset` should have a corresponding
        element in the index of `split_table`. Opposite is not necessary, i.e.
        `split_table` can have elements that are absent in `dataset` (for
        example when subsampling).

        Returns
        -------
        split: Split
            Split object with dataset split according to the `split_table`.

        """
        train_dataset = dataset[split_table == SplitFlag.TRAIN]
        train_dataset.name += '/train'
        val_dataset = dataset[split_table == SplitFlag.VAL]
        val_dataset.name += '/val'
        test_dataset = dataset[split_table == SplitFlag.TEST]
        test_dataset.name += '/test'

        return cls(train=train_dataset, val=val_dataset, test=test_dataset)

    def to_split_table(self, sort_index: bool = True) -> pd.Series:
        """ Generate a split table from a Split object

        Parameters
        ----------
        sort_index: bool, optional
            If True (default) the index is be sorted before returning the split
            table. Consider setting to False for huge data sets.

        Returns
        -------
        split_table: pd.Series
            Series with int flags indicating to which dataset each row belongs.
            See `SplitFlag` enum for exact meanings of the flags.
        """
        ind_train = pd.Series(SplitFlag.TRAIN, index=self.train.x.index)
        ind_val = pd.Series(SplitFlag.VAL, index=self.val.x.index)
        ind_test = pd.Series(SplitFlag.TEST, index=self.test.x.index)

        split_table = pd.concat([ind_train, ind_test, ind_val])

        if sort_index:
            split_table.sort_index(inplace=True)

        return split_table



# Copyright 2020 (c) Netguru S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import warnings
from typing import Optional, List, Callable, Tuple, Union

import optuna
import skopt
from optuna import Study
from optuna.integration import SkoptSampler
from optuna.samplers import RandomSampler, TPESampler, BaseSampler

from opoca.hyperparam_search.objective import Objective
from opoca.hyperparam_search.utils import OPTUNA_DB


class POCStudy(Study):
    """ Performs parameters optimization and takes care of logging.

    Parameters
    ----------
    objective: Objective
        Objective object containing the function that should be optimized.
    direction: str, optional
        One of ['maximize', 'minimize']. Use 'maximize' for accuracy, f1 etc,
         use 'minimize' for error rates, loss etc.
    algorithm: str or skopt.Optimizer, optional
        Algorithm used to find best hyperparameters.
        There are following options:
        * 'TPE' - Tree Parzen Estimator, default and decent one. It handles
        well conditional hyperparameter spaces.
        * 'GP' - Gaussian Process, most famous and sophisticated, but cannot
        handle conditional hyperparameter spaces. It falls back to the random
        search for parameters that are conditioned on other parameters.
        * 'RF' - Use Random Forest Regressor as a surrogate model. Rarely used.
        * 'random' - simple and can be parallelized without any synchronization
            between instances.
        * If more flexibility is needed, provide directly a skopt.Optimizer
        instance.

    pruner: subclass of BasePruner or None, optional
        When using a k-fold cross validation, after each fold, the score is
        passed to the pruner. If it looks unpromising, the trial is stopped
        without evaluation of other folds. By default, a median pruner is used.
        It prunes if the trial's best intermediate result is worse than median
        of intermediate results of previous trials at the same step.
        If None, a dummy pruner that never prunes will be used.

    Notes
    -----
    TPE, GP, RF should be rather used with num_jobs=1.

    References
    ----------
    __init__ was inspired by `optuna.create_study`.
    """

    # A sentinel that allows specifying meaningfully None as the pruner.
    DEFAULT_PRUNER = object()

    def __init__(self,
                 objective: Objective,
                 direction: str = 'maximize',
                 algorithm: Union[str, skopt.Optimizer] = 'TPE',
                 pruner: Optional[optuna.pruners.BasePruner] = DEFAULT_PRUNER):

        storage = os.environ.get(OPTUNA_DB, None)
        if storage is not None:
            print('Using RDB for storing trial results.')
            storage = optuna.storages.RDBStorage(url=storage)
            version = storage.get_current_version()
            print(f'Successfully connected to the database (version: {version}).')

        study_name = objective.name
        assert study_name is not None
        storage = optuna.storages.get_storage(storage)
        try:
            study_id = storage.create_new_study(study_name)
            print(f"Created a new study with name '{study_name}'")
        except optuna.exceptions.DuplicatedStudyError:
            study_id = storage.get_study_id_from_name(study_name)
            print(f"Using an existing study with name '{study_name}' instead of creating a new one.")

        self.sampler_name, sampler = self.__create_sampler(algorithm)

        if pruner is None:
            # In optuna, pruner=None leads to usage of MedianPruner, which is
            # counter-intuitive. None is None. NopPruner never prunes.
            pruner = optuna.pruners.NopPruner()
        elif pruner is self.DEFAULT_PRUNER:
            # We use here value 15 instead of default 5, because for TPE we
            # need gather some initial knowledge about the search space.
            # You can provide directly a pruner instance if you need another value.
            pruner = optuna.pruners.MedianPruner(n_startup_trials=15)

        super().__init__(study_name=study_name, storage=storage, sampler=sampler, pruner=pruner)

        if direction == "minimize":
            _direction = optuna.structs.StudyDirection.MINIMIZE
        elif direction == "maximize":
            _direction = optuna.structs.StudyDirection.MAXIMIZE
        else:
            raise ValueError("Please set either 'minimize' or 'maximize' to direction.")

        self._storage.set_study_direction(study_id, _direction)

        self.objective = objective

    def __create_sampler(self, algorithm: Union[str, skopt.Optimizer]) -> (str, BaseSampler):
        """ Create sampler from Optimizer instance or string

        Parameters
        ----------
        sampler_name: skopt.Optimizer or str
            Passed from constructor - see class doc.

        Returns
        -------
        Tuple: sampler_name and an instance of optuna BaseSampler subclass
        """
        if isinstance(algorithm, skopt.Optimizer):
            sampler = algorithm
            sampler_name = algorithm.__class__.__name__
        elif algorithm == 'TPE':
            sampler = TPESampler()
            sampler_name = algorithm
        elif algorithm == 'GP':
            sampler = SkoptSampler()
            sampler_name = algorithm
            # `n_startup_trials` of optuna SkoptSampler should not be changed.
            # The internal scikit-optimize sampler has `n_initial_points=10`.
            # If needed you can change it using `skopt_kwargs`.
        elif algorithm == 'RF':
            skopt_kwargs = {'base_estimator': 'RF'}
            sampler = SkoptSampler(skopt_kwargs=skopt_kwargs)
            sampler_name = algorithm
        elif algorithm == 'random':
            sampler = RandomSampler()
            sampler_name = algorithm
        else:
            raise ValueError(f'Unknown hyperparameter search algorithm: {algorithm}. '
                             f'Use one of those: "TPE", "GP", "RF" or "random".')

        return sampler_name, sampler

    def optimize(  # pylint: disable=arguments-differ
            self,
            func: None = None,
            num_trials: Optional[int] = 50,
            timeout_seconds: Optional[float] = None,
            num_jobs: int = 1,
            catch: Tuple = (),
            callbacks: Optional[List[Callable]] = None,
            gc_after_trial: bool = True,
            show_progress_bar: bool = False):
        """ Optimize the objective function

        Parameters
        ----------
        func: None
            Ignored, only for compatibility with base class.
        other parameters:
             -> see optuna.Study.optimize for the parameters description
        """
        if func is not None:
            warnings.warn('`func` was provided but will ignored. The objective '
                          'that was provided in the constructor will be used. '
                          'The argument is present only to provide compatibility '
                          'with the  base class.')

        if self.sampler_name in ["TPE", "GP", "RF"] and num_jobs != 1:
            warnings.warn(f'The chosen algorithm ({self.sampler_name}) is not '
                          f'parallelizable and may behave strange if run in '
                          f'multiple instances. For example, it may visit the '
                          f'same point multiple times.')

        return super().optimize(
            self.objective,
            n_trials=num_trials,
            timeout=timeout_seconds,
            n_jobs=num_jobs,
            catch=catch,
            callbacks=callbacks,
            gc_after_trial=gc_after_trial,
            show_progress_bar=show_progress_bar)



# Copyright 2020 (c) Netguru S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import os
from typing import Dict

import pandas as pd
from dotenv import load_dotenv
from simple_salesforce import Salesforce

load_dotenv()


def get_data_frame_from_salesforce(sf_instance: Salesforce, table_name: str) -> pd.DataFrame:
    """
    Queries for data in SalesForce for a given table.
    Parameters
    ----------
    sf_instance
        Instance of SalesForce connector
    table_name
        Name of table in SalesForce
    Returns
    -------
    Table rows as DataFrame
    """
    desc = sf_instance.__getattr__(table_name).describe()

    field_names = [field['name'] for field in desc['fields']]
    soql = "SELECT {} FROM {}".format(','.join(field_names), table_name)
    results = sf_instance.query_all(soql)

    return pd.DataFrame().from_dict(results['records'])


class SFCache:
    """
    Simple cache for SalesForce connector. Assumes credentials to salesforce are stored in environment variables:
    * `SF_{prefix}_username`
    * `SF_{prefix}_password`
    * `SF_{prefix}_security_token`

    `prefix` can be `{'staging', 'production'}`.

    Parameters
    ----------
    data_root
        Path to local data storage directory
    source
        'production' or 'staging'
    """
    def __init__(self, data_root: str, source: str = "production"):
        self.cache_dir_path = os.path.join(data_root, f"sf_cache_{source}")
        os.makedirs(self.cache_dir_path, exist_ok=True)
        self.kwargs = self.__get_sf_kwargs(source)
        self.sf_instance = None

    @staticmethod
    def __get_sf_kwargs(source: str) -> Dict[str, str]:
        """
        Prepares kwargs for SalesForce connector
        """
        if source == "staging":
            kwargs = dict(domain="test")
        elif source == "production":
            kwargs = dict()
        else:
            raise ValueError("Source should be either production or staging")

        password = os.environ[f"SF_{source}_password"]
        username = os.environ[f"SF_{source}_username"]
        security_token = os.environ[f"SF_{source}_security_token"]

        credentials_dict = dict(password=password,
                                username=username,
                                security_token=security_token)

        kwargs.update(credentials_dict)

        return kwargs

    def load(self, table_name: str, refresh_cache: bool = False) -> pd.DataFrame:
        """
        Loads data from SalesForce and saves in local file. If data already exists and `force` flag is off, then it
        loads data from local file. If `force` is on, then it always loads data from SalesForce
        Parameters
        ----------
        table_name
            Name of table that should be loaded
        refresh_cache
            Enforces to not use cached local file
        Returns
        -------
        Table as data frame
        """
        logging.info(f"Loading table {table_name}")
        table_csv_path = os.path.join(self.cache_dir_path, f'{table_name}.csv')
        if os.path.exists(table_csv_path) and not refresh_cache:
            logging.debug(f"Loading data from cache located in {table_csv_path}")
            df = pd.read_csv(table_csv_path)
        else:
            logging.debug("Downloading data from Salesforce")
            if self.sf_instance is None:
                self.sf_instance = Salesforce(**self.kwargs)

            df = get_data_frame_from_salesforce(self.sf_instance, table_name)
            df.to_csv(table_csv_path)
        return df



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['opoca',
 'opoca.data',
 'opoca.data.integrations',
 'opoca.evaluation',
 'opoca.features',
 'opoca.hyperparam_search',
 'opoca.models',
 'opoca.trainers']

package_data = \
{'': ['*']}

install_requires = \
['Cython>=0.29.20,<0.30.0',
 'MonthDelta>=0.9.1,<0.10.0',
 'category_encoders>=2.2.2,<3.0.0',
 'click>=7.1.2,<8.0.0',
 'dynaconf>=2.2.3,<3.0.0',
 'google-cloud-storage>=1.29.0,<2.0.0',
 'horology>=1.1.0,<2.0.0',
 'imbalanced-learn>=0.7.0,<0.8.0',
 'ipython>=7.16.1,<8.0.0',
 'joblib>=0.15.1,<0.16.0',
 'kubernetes>=11.0.0,<12.0.0',
 'matplotlib>=3.2.2,<4.0.0',
 'mlflow>=1.9.1,<2.0.0',
 'numpy>=1.19.0,<2.0.0',
 'optuna>=1.5.0,<2.0.0',
 'pandas-profiling>=2.8.0,<3.0.0',
 'pandas>=1.0.5,<2.0.0',
 'plotly>=4.8.1,<5.0.0',
 'pyarrow>=0.17.1,<0.18.0',
 'python-dateutil==2.8.0',
 'python-dotenv>=0.13.0,<0.14.0',
 'quilt3>=3.1.14,<4.0.0',
 'scikit-learn>=0.23.1,<0.24.0',
 'scikit-optimize>=0.7.4,<0.8.0',
 'scipy>=1.5.0,<2.0.0',
 'seaborn>=0.10.1,<0.11.0',
 'simple-salesforce>=1.1.0,<2.0.0',
 'tqdm>=4.46.1,<5.0.0',
 'urllib3==1.24.3',
 'xgboost>=1.1.1,<2.0.0']

setup_kwargs = {
    'name': 'opoca',
    'version': '0.1.0',
    'description': 'Opoca library aims to drastically speed up producing proof of concepts (PoC) for machine learning projects.',
    'long_description': None,
    'author': 'Apollo Team',
    'author_email': 'ml-team@netguru.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# Copyright 2020 (c) Netguru S.A.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
useful links
------------
https://towardsdatascience.com/adversarial-validation-ca69303543cd
https://www.kaggle.com/konradb/adversarial-validation-and-other-scary-terms
"""

from math import sqrt
from typing import Tuple
from warnings import warn

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from horology import timed
from pandas import get_dummies
from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import roc_auc_score
from sklearn.model_selection import StratifiedKFold

RANDOM_CLASSIFIER_AUROC = 0.5
DECENT_CLASSIFIER_AUROC = 0.8
SIGNIFICANT_Z_SCORE = 2
EPSILON = 1.e-6


def drop_uncommon_columns(df_a: pd.DataFrame, df_b: pd.DataFrame) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """ Drops columns that are not in both DataFrames.

    Parameters
    ----------
    df_a, df_b: pd.DataFrame
        DataFrames that should have columns synchronised.

    Returns
    -------
    df_a, df_b: pd.DataFrame
        DataFrames with the same columns.

    """
    common_columns = set(df_a.columns).intersection(set(df_b.columns))

    return df_a.loc[:, common_columns], df_b.loc[:, common_columns]


def transform_to_numeric_form(df: pd.DataFrame) -> pd.DataFrame:
    """ Transform all data to be in numeric form

    A few steps are performed:
    1. Remove timestamp from features if any, because it can be easily used to
    distinguish the train set from the test set in case of time split.
    2. Category/object columns are transformed to one-hot encoded (dummy)
    variables. It deals well with <NA>.
    3. <NA> are filled with median.

    Those steps may not be suitable for every dataset.
    They are designed to be as general as possible and to work with every data.
    Consider more hand-crafted preprocessing if any warning appear.

    Parameters
    ----------
    df: pd.DataFrame
        Initially preprocessed data that should be transformed.

    Returns
    -------
    df: pd.DataFrame
        A DataFrame with all transformation performed.

    """
    STANDARD_WARNING = '\nKeep in mind that the dataset should be initially preprocessed.'

    # remove timestamp
    columns_before = set(df.columns)
    df = df.select_dtypes(exclude='datetime')
    columns_after = set(df.columns)
    if columns_after != columns_before:
        warn('Columns with dtype=datetime were removed. You may want to use '
             'timedelta features instead.' + STANDARD_WARNING)

    # encode all categorical features as one-hot
    num_cols_before = len(df.columns)
    df = get_dummies(df)
    num_cols_after = len(df.columns)
    if num_cols_after > 10 * num_cols_before:
        warn(f'Number of columns after dummy encoding changed from '
             f'{num_cols_before} to {num_cols_after}. It can mean that:\n'
             f'(1) you have category type columns with a lof of unique variables (this is fine), or \n'
             f'(2) a numeric column was treated as category-like and every value is one-hot encoded (this is wrong).'
             f'{STANDARD_WARNING}')

    # fill <NA>
    if df.isna().any().any():
        df.fillna(df.median(), inplace=True)
        warn('Missing values are filled with medians.' + STANDARD_WARNING)

    return df


def create_adversarial_validation_data(train: pd.DataFrame, test: pd.DataFrame) \
        -> Tuple[pd.DataFrame, pd.DataFrame]:
    """ Creates adversarial validation dataset, i.e. binary classification data
    with test set being the positive class. It performs some necessary
    preprocessing - drops uncommon columns and transforms all data to numeric
    forms.


    Parameters
    ----------
    train: pd.DataFrame
        train data set
    test: pd.DataFrame
        test data set

    Returns
    -------
    x: pd.DataFrame
        processed features. For details see doc string of `fix_everything`
    y: pd.DataFrame
        binary targets: 0 for train, 1 for test

    See also
    --------
    `drop_uncommon_columns` and `transform_to_numeric_form` docstrings

    """
    train, test = drop_uncommon_columns(train, test)
    train.loc[:, 'is_test'] = 0
    test.loc[:, 'is_test'] = 1

    data = pd.concat([train, test], keys=['train', 'test'])
    data = transform_to_numeric_form(data)
    y = data.pop('is_test').astype('uint8')

    return data, y


def interpret(roc_auc: float, z_score: float) -> str:
    """ Subjective interpretation of adversarial validation.

    Parameters
    ----------
    roc_auc: float
        roc_auc score of train vs test classifier
    z_score: float
        how far are we from random classifier, measured in standard deviations

    Returns
    -------
    Subjective opinion about the situation.

    """
    if z_score < 1 - SIGNIFICANT_Z_SCORE:
        return 'Classifier did not converge on the data.'

    if z_score < SIGNIFICANT_Z_SCORE:
        if roc_auc < DECENT_CLASSIFIER_AUROC:
            return 'The test set is probably from the same distribution as the train set.'
        else:
            # This is a strange situation: roc is high, but its variance is
            # also very high yielding small z_score
            return 'Analysis is non-conclusive. Possibly more data would help.'
    else:
        if roc_auc < DECENT_CLASSIFIER_AUROC:
            return 'Test and train set differ.'
        else:
            return 'Test and train set substantially differ.'


class AdversarialValidator:
    """ Performs automagic analysis of train/test set distributions. It tries
    to fit a classifier that can distinguish if a given example is from test or
    train set. High classifier performance means that those sets differ. See
    `do_analysis` method for more details.

    Parameters
    ----------
    train: pd.DataFrame
        train data set
    test: pd.DataFrame
        test data set

    Example
    -------
    av = AdversarialValidator(train, test)
    av.do_analysis()
    av.plot_features_importances()

    """

    def __init__(self, train: pd.DataFrame, test: pd.DataFrame):
        self.x, self.y = create_adversarial_validation_data(train, test)
        self.classifiers_: list
        self.feat_importances_: pd.Series

    def _roc_score(self, num_cv_splits: int = 5) -> Tuple[float, float]:
        """ Calculates mean area under receiver operating curve for test/train
        classifier. Mean is calculated over `num_cv_splits` folds of
        RandomForestClassifier trainings.

        This method also populates `self.feat_importances` with mean feature
        importances from each fold. Keep in mind that feature importances may
        be misleading for unbalanced data.

        Parameters
        ----------
        num_cv_splits: int, optional
            How many cross validation splits should be done.

        Returns
        -------
        roc_mean: float
            average values of auc roc score
        roc_std: float
            standard deviation of auc roc score

        """
        self.classifiers_ = []
        rocs = []
        cross_validation = StratifiedKFold(n_splits=num_cv_splits, shuffle=True)

        # this highly confusing heuristics limits number of trees in random
        # forest for small sets
        num_estimators = int(max(3, min(100, 0.5 * sqrt(len(self.y)))))

        for train_index, val_index in cross_validation.split(self.x, self.y):
            x_train, x_val = self.x.iloc[train_index], self.x.iloc[val_index]
            y_train, y_val = self.y.iloc[train_index], self.y.iloc[val_index]

            classifier = RandomForestClassifier(num_estimators, n_jobs=-1)
            classifier.fit(x_train, y_train)
            y_pred = classifier.predict_proba(x_val)[:, 1]
            self.classifiers_.append(classifier)

            roc = roc_auc_score(y_val, y_pred)
            rocs.append(roc)

        roc_mean = np.mean(rocs)
        roc_std = np.std(rocs)

        cv_feat_importances = pd.concat(
            [pd.Series(cl.feature_importances_, index=self.x.columns)
             for cl in self.classifiers_])

        self.feat_importances_ = \
            cv_feat_importances.groupby(cv_feat_importances.index).mean()

        return roc_mean, roc_std

    def do_analysis(self):
        """ Checks if a classifier can distinguish between train and test sets.

        It checks two values:
        - classifier auc roc scores. It should be 0.5 when it is impossible to
        distinguish sets. Values close to 1 indicate that datasets differ
        substantially.
        - z_score, which can be interpreted as distance (measured in standard
        deviations) between random classifier expected roc and actual roc.
        High z_score mean that the difference between sets is not accidental,
        but statistically significant.

        In other words, the roc can be interpreted as a measure of how the sets
        differ and z_score tells us how sure we are about it.

        See z_score explanation here:
        https://www.statisticshowto.datasciencecentral.com/probability-and-statistics/z-score/

        Returns
        -------
        roc_mean: float
            average values of auc roc score
        z_score: float
            z score as explained above

        """
        roc_mean, roc_std = self._roc_score()
        z_score = (roc_mean - RANDOM_CLASSIFIER_AUROC) / (roc_std + EPSILON)

        print(f'mean roc: {roc_mean:.2f} (std={roc_std:.2f}) '
              f'z_score: {z_score:.2f} - {interpret(roc_mean, z_score)}')

        return roc_mean, z_score

    def drop_most_important_feature(self):
        """ Drops a columns that contributes the most to classification accuracy.

        Remarks
        -------
        Feature importances may be misleading, especially for unbalanced data.
        It is also biased towards features with high cardinality (for example
        float vs binary). Treat feature importances rather as a hint on which
        features you should inspect further.

        Returns
        -------
        most_important: str
            name of the most significant feature
        importance: float
            its importance from sk-learn

        """
        if not hasattr(self, 'feat_importances_'):
            raise AttributeError('Cannot drop a feature before analysis is performed.')

        most_important = self.feat_importances_.idxmax()
        importance = self.feat_importances_.max()
        self.x.pop(most_important)

        del self.feat_importances_

        return most_important, importance

    def plot_features_importances(self, num_largest: int = 12):
        """ Plots `num_largest` most important features as a horizontal bar plot

        Parameters
        ----------
        num_largest: int, optional
            How many features should be on the plot

        """
        if not hasattr(self, 'feat_importances_'):
            raise AttributeError('Cannot plot feature importances before analysis is performed.')
        self.feat_importances_.nlargest(num_largest).plot(kind='barh')
        plt.show()


@timed
def find_most_divergent_features(train: pd.DataFrame, test: pd.DataFrame,
                                 num_features_to_drop: int = 3,
                                 show_plots: bool = True) -> list:
    """ Use AdversarialValidator to find features that may have different
    distributions in test and train sets

    Parameters
    ----------
    train: pd.DataFrame
        train data set
    test: pd.DataFrame
        test data set
    num_features_to_drop: int, optional
        how many most relevant features should be dropped
    show_plots: bool, optional
        if bar plots should be generated

    Returns
    -------
    dropped_features: list
        list of the most divergent features

    """
    dropped_features = []
    adv_validator = AdversarialValidator(train, test)

    initial_roc, initial_z_score = adv_validator.do_analysis()
    if show_plots:
        adv_validator.plot_features_importances()

    for _ in range(num_features_to_drop):
        most_important, importance = adv_validator.drop_most_important_feature()
        dropped_features.append(most_important)

        print(f'`{most_important}` (importance={importance:.3f}) is the most '
              f'important feature allowing distinguish the train set from the '
              f'test set. It was removed. \n')

        final_roc, final_z_score = adv_validator.do_analysis()
        if show_plots:
            adv_validator.plot_features_importances()

    print(f'\n\n\n'
          f'After dropping {dropped_features}\n'
          f'roc {initial_roc:.2f} -> {final_roc:.2f}\n'
          f'z_score {initial_z_score:.2f} -> {final_z_score:.2f}\n')

    return dropped_features
